import copy
import glob
import os
import sys
import pprint
from itertools import groupby
from textwrap import wrap
import numpy as np
import pandas as pd
import pylab as plt
import tqdm
from . import plots_line
from .. import haven_utils as hu


def get_images(exp_list, savedir_base, n_exps=20, n_images=1,
                   figsize=(12,12), legend_list=None,
                   dirname='images', verbose=True):
    """[summary]
    
    Parameters
    ----------
    exp_list : list
        A list of experiments, each defines a single set of hyper-parameters
    savedir_base : str
        A directory where experiments are saved
    n_exps : int, optional
        [description], by default 3
    n_images : int, optional
        [description], by default 1
    height : int, optional
        [description], by default 12
    width : int, optional
        [description], by default 12
    legend_list : [type], optional
        [description], by default None
    dirname : str, optional
        [description], by default 'images'

    Returns
    -------
    fig_list : list
        a list of pylab figures

    Example
    -------
    >>> from haven import haven_results as hr
    >>> savedir_base='../results/isps/'
    >>> exp_list = hr.get_exp_list(savedir_base=savedir_base, 
    >>>                            filterby_list=[{'sampler':{'train':'basic'}}])
    >>> hr.get_images(exp_list, savedir_base=savedir_base)
    """
    fig_list = []
    exp_count = 0
    for k, exp_dict in enumerate(exp_list):
        
        if exp_count >= n_exps:
            if verbose:
                print('displayed %d/%d experiment images' % (k, n_exps))
            break

        result_dict = {}
        

        exp_id = hu.hash_dict(exp_dict)
        result_dict['exp_id'] = exp_id
        if verbose:
            print('Displaying Images for Exp:', exp_id)
        savedir = os.path.join(savedir_base, exp_id)

        base_dir = os.path.join(savedir, dirname)
        img_list = glob.glob(os.path.join(base_dir, '*.jpg'))
        img_list += glob.glob(os.path.join(base_dir, '*.png'))

        img_list.sort(key=os.path.getmtime)
        img_list = img_list[::-1]
        img_list = img_list[:n_images]

        if len(img_list) == 0:
            if verbose:
                print('no images in %s' % base_dir)
            continue

        ncols = len(img_list)
        # ncols = len(exp_configs)
        nrows = 1
        # from IPython.display import display
        # display('%s' % ("="*50))
        result_dict = {'exp_id': exp_id}
        result_dict.update(copy.deepcopy(exp_dict))
        score_list_path = os.path.join(savedir, 'score_list.pkl')
        if os.path.exists(score_list_path):
            score_list = hu.load_pkl(score_list_path)
            result_dict.update(score_list[-1])
        # display(pd.Series(result_dict))
        if legend_list is not None:
            label = plots_line.get_label(legend_list, exp_dict, show_key=True)
        else:
            label = exp_id
        
        if 'epoch' in result_dict:
            label += "_epoch:%d" % result_dict['epoch']
        # if legend_list is None:
        #     label = hu.hash_dict(exp_dict)
        # else:
        #     label = '-'.join(['%s:%s' % (k, str(result_dict.get(k))) for
        #                         k in legend_list])

        for i in range(ncols):
            img_fname = os.path.split(img_list[i])[-1]
            fig = plt.figure(figsize=figsize)
            try:
                img = plt.imread(img_list[i])
                plt.imshow(img)
                plt.title('%s - %s\n%s' %
                                (exp_id, img_fname, label),
                                fontsize=28)

                plt.axis('off')
                plt.tight_layout()
                fig_list += [fig]

            except:
                print('skipping - %s, image corrupted' % img_fname)
            
            
        exp_count += 1
    
    return fig_list


import contextlib
import copy
import hashlib
import itertools
import json


import pprint
import os
import pickle
import shlex
import subprocess
import threading
import time
import numpy as np
import pylab as plt
import torch

def save_json(fname, data, makedirs=True):
    """Save data into a json file.

    Parameters
    ----------
    fname : str
        Name of the json file
    data : [type]
        Data to save into the json file
    makedirs : bool, optional
        If enabled creates the folder for saving the file, by default True
    """
    dirname = os.path.dirname(fname)
    if makedirs and dirname != '':
        os.makedirs(dirname, exist_ok=True)
    with open(fname, "w") as json_file:
        json.dump(data, json_file, indent=4, sort_keys=True)


def load_mat(fname):
    """Load a matlab file.

    Parameters
    ----------
    fname : str
        File name

    Returns
    -------
    dict
        Dictionary with the loaded data
    """
    import scipy.io as io
    return io.loadmat(fname)


def load_json(fname, decode=None):  # TODO: decode???
    """Load a json file.

    Parameters
    ----------
    fname : str
        Name of the file
    decode : [type], optional
        [description], by default None

    Returns
    -------
    [type]
        Content of the file
    """
    with open(fname, "r") as json_file:
        d = json.load(json_file)

    return d


def read_text(fname):
    """Loads the content of a text file.

    Parameters
    ----------
    fname : str
        File name

    Returns
    -------
    list
        Content of the file. List containing the lines of the file
    """
    with open(fname, "r", encoding="utf-8", errors='replace') as f:
        lines = f.readlines()
    return lines


def load_pkl(fname):
    """Load the content of a pkl file.

    Parameters
    ----------
    fname : str
        File name

    Returns
    -------
    [type]
        Content of the file
    """
    with open(fname, "rb") as f:
        return pickle.load(f)


def save_pkl(fname, data, with_rename=True, makedirs=True):
    """Save data in pkl format.

    Parameters
    ----------
    fname : str
        File name
    data : [type]
        Data to save in the file
    with_rename : bool, optional
        [description], by default True
    makedirs : bool, optional
        If enabled creates the folder for saving the file, by default True
    """
    # Create folder
    dirname = os.path.dirname(fname)
    if makedirs and dirname != '':
        os.makedirs(dirname, exist_ok=True)

    # Save file
    if with_rename:
        fname_tmp = fname + "_tmp.pth"
        with open(fname_tmp, "wb") as f:
            pickle.dump(data, f)
        if os.path.exists(fname):
            os.remove(fname)
        os.rename(fname_tmp, fname)
    else:
        with open(fname, "wb") as f:
            pickle.dump(data, f)


import shutil
import os, torch

from . import haven_utils as hu



def delete_experiment(savedir, backup_flag=False):
    """Delete an experiment. If the backup_flag is true it moves the experiment
    to the delete folder.
    
    Parameters
    ----------
    savedir : str
        Directory of the experiment
    backup_flag : bool, optional
        If true, instead of deleted is moved to delete folder, by default False
    """
    # get experiment id
    exp_id = os.path.split(savedir)[-1]
    assert(len(exp_id) == 32)

    # get paths
    savedir_base = os.path.dirname(savedir)
    savedir = os.path.join(savedir_base, exp_id)

    if backup_flag:
        # create 'deleted' folder 
        dst = os.path.join(savedir_base, 'deleted', exp_id)
        os.makedirs(dst, exist_ok=True)

        if os.path.exists(dst):
            shutil.rmtree(dst)
    
    if os.path.exists(savedir):
        if backup_flag:
            # moves folder to 'deleted'
            shutil.move(savedir, dst)
        else:
            # delete experiment folder 
            shutil.rmtree(savedir)

    # make sure the experiment doesn't exist anymore
    assert(not os.path.exists(savedir))

def load_checkpoint(exp_dict, savedir_base, fname='model_best.pth'):
    path = os.path.join(savedir_base, hu.hash_dict(exp_dict), fname)
    return torch.load(path)

def delete_and_backup_experiment(savedir):
    """Delete an experiment and make a backup (Movo to the trash)
    
    Parameters
    ----------
    savedir : str
        Directory of the experiment
    """
    # delete and backup experiment
    delete_experiment(savedir, backup_flag=True)

def get_savedir(exp_dict, savedir_base):
    """[summary]
    
    Parameters
    ----------
    exp_dict : dict
        Dictionary describing the hyperparameters of an experiment
    savedir_base : str
        Directory where the experiments are saved
    
    Returns
    -------
    str
        Directory of the experiment
    """
    # get experiment savedir
    exp_id = hu.hash_dict(exp_dict)
    savedir = os.path.join(savedir_base, exp_id)
    return savedir


def copy_checkpoints(savedir_base, filterby_list, hparam, hparam_new):
    """
    """
    # get experiment savedir
    import copy, tqdm
    from . import haven_utils as hr

    exp_list = hr.get_exp_list(savedir_base=savedir_base)
    exp_list_new = hr.filter_exp_list(exp_list, filterby_list=filterby_list)
    exp_list_out = copy.deepcopy(exp_list_new)
    for exp_dict in exp_list_out:
        exp_dict[hparam_new] =  exp_dict[hparam]
        del exp_dict[hparam]

    for e1, e2 in tqdm.tqdm(zip(exp_list_new, exp_list_out)):
        h1 = hu.hash_dict(e1)
        s1 = os.path.join(args.savedir_base, h1)

        h2 = hu.hash_dict(e2)
        s2 = os.path.join(args.savedir_base, h2)

        # copy exp dict
        os.makedirs(s2, exist_ok=True)
        e2_fname = os.path.join(s2, 'exp_dict.json')
        hu.save_json(e2_fname, e2)

        for fname in ['score_list.pkl', 'borgy_dict.json', 'job_dict.json', 'logs.txt', 'err.txt']:
            
            # copy score list
            s1_fname = os.path.join(s1, fname)
            s2_fname = os.path.join(s2, fname)
            # assert(not os.path.exists(s_list_fname))
            if os.path.exists(s1_fname):
                if '.json' in fname:
                    hu.save_json(s2_fname, hu.load_json(s1_fname))
                elif '.pkl' in fname:
                    hu.save_pkl(s2_fname, hu.load_pkl(s1_fname))
                elif '.txt' in fname:
                    hu.save_txt(s2_fname, hu.load_txt(s1_fname)[-5:])


import copy
import glob
import os
import sys
import pprint
from itertools import groupby
from textwrap import wrap
import numpy as np
import pandas as pd
import pylab as plt
import tqdm

from .. import haven_jobs as hjb
from .. import haven_utils as hu
from .. import haven_share as hd

def get_score_df(exp_list, savedir_base, filterby_list=None, columns=None,
                 score_columns=None,
                 verbose=True, wrap_size=8, hparam_diff=0, flatten_columns=True,
                 show_meta=True, show_max_min=True, add_prefix=False,
                 score_list_name='score_list.pkl', in_latex_format=False, avg_across=None,
                 return_columns=False):
    """Get a table showing the scores for the given list of experiments 

    Parameters
    ----------
    exp_list : list
        A list of experiments, each defines a single set of hyper-parameters
    columns : list, optional
        a list of columns you would like to display, by default None
    savedir_base : str, optional
        A directory where experiments are saved
        
    Returns
    -------
    DataFrame
        a dataframe showing the scores obtained by the experiments

    Example
    -------
    >>> from haven import haven_results as hr
    >>> savedir_base='../results/isps/'
    >>> exp_list = hr.get_exp_list(savedir_base=savedir_base, 
    >>>                            filterby_list=[{'sampler':{'train':'basic'}}])
    >>> df = hr.get_score_df(exp_list, savedir_base=savedir_base, columns=['train_loss', 'exp_id'])
    >>> print(df)
    """
    if len(exp_list) == 0:
        if verbose:
            print('exp_list is empty...')
        if return_columns:
            return pd.DataFrame([]), [], []
        else:
            return pd.DataFrame([])
    exp_list = hu.filter_exp_list(exp_list, filterby_list, savedir_base=savedir_base, verbose=verbose)

    # aggregate results
    hparam_list = set()
    result_list = []
    for exp_dict in exp_list:
        result_dict = {}
        exp_id = hu.hash_dict(exp_dict)
        if avg_across is not None:
            tmp_dict = copy.deepcopy(exp_dict)
            del tmp_dict[avg_across]
            result_dict['_' + avg_across] = hu.hash_dict(tmp_dict)

        savedir = os.path.join(savedir_base, exp_id)
        score_list_fname = os.path.join(savedir, score_list_name)
        exp_dict_fname = os.path.join(savedir, "exp_dict.json")

        
        if flatten_columns:
            exp_dict_flat = hu.flatten_column(exp_dict)
        else:
            exp_dict_flat = exp_dict
        
        hparam_columns = columns or list(exp_dict_flat.keys())
        for hc in hparam_columns:
            hparam_list.add(hc)
        for k in hparam_columns:
            if add_prefix:
                k_new = "(hparam) " + k
            else:
                k_new = k
            if k not in exp_dict_flat:
                continue
            result_dict[k_new] = exp_dict_flat[k]

        if os.path.exists(score_list_fname) and show_meta:
            result_dict['started_at'] = hu.time_to_montreal(exp_dict_fname)
            result_dict['creation_time'] = os.path.getctime(exp_dict_fname)
        else:
            result_dict['creation_time'] = -1

        if show_meta:
            result_dict["exp_id"] = exp_id

        # hparam_columns = [k for k in result_dict.keys() if k not in ['creation_time']]

        if not os.path.exists(score_list_fname):
            if verbose:
                print('%s: %s is missing' % (exp_id, score_list_name))
            
        else:
            score_list = hu.load_pkl(score_list_fname)
            score_df = pd.DataFrame(score_list)
            metric_columns = score_columns or score_df.columns
            if len(score_list):
                for k in metric_columns:
                    if k not in score_df.columns:
                        continue
                    v = np.array(score_df[k])
                    if 'float' in str(v.dtype):
                        v = v[~np.isnan(v)]

                    if len(v):
                        if add_prefix:
                            k_new = "(metric) " +k
                        else:
                            k_new = k

                        if "float" in str(v.dtype):
                            result_dict[k_new] = v[-1]
                            if show_max_min:
                                result_dict[k_new+' (max)'] = v.max()
                                result_dict[k_new+' (min)'] = v.min()
                                
                        else:
                            result_dict[k_new] = v[-1]
        

        result_list += [result_dict]
    
    # create table
    df = pd.DataFrame(result_list)
    hparam_columns = list(hparam_list)
    metric_columns = [c for c in df.columns if c not in hparam_columns + ['creation_time']]
    # print(avg_across)
    if avg_across is not None:
        df_avg = df.groupby('_' + avg_across)
        df_tmp = df_avg[hparam_columns].first().join(df_avg[metric_columns].agg([np.mean, np.std]))
        
        if ('_' + avg_across) in df_tmp.columns:
            del df_tmp['_' + avg_across]
        
        df_tmp = df_tmp.reset_index()

        if ('_' + avg_across) in df_tmp.columns:
            del df_tmp['_' + avg_across]
            
        df = df_tmp 
        
    if 'creation_time' in df.columns:
        df = df.sort_values(by='creation_time' )
        del df['creation_time']

    # wrap text for prettiness
    df = hu.pretty_print_df(df)

    if hparam_diff > 0 and len(df) > 1:
        cols = hu.get_diff_columns(df, min_threshold=hparam_diff, max_threshold='auto')
        df = df[cols]

    df =  hu.sort_df_columns(df, also_first=hparam_columns)
    if in_latex_format:
        return df.to_latex(index=False)

    if return_columns:
        return df, sorted(hparam_columns), sorted(metric_columns)
    return df

def get_score_lists(exp_list, savedir_base, filterby_list=None, verbose=True,
                    score_list_name='score_list.pkl'):
    """[summary]
    
    Parameters
    ----------
    exp_list : [type]
        A list of experiments, each defines a single set of hyper-parameters
    savedir_base : [type]
        [description]
    filterby_list : [type], optional
        [description], by default None
    
    Returns
    -------
    [type]
        [description]
    Example
    -------
    >>> from haven import haven_results as hr
    >>> savedir_base='../results/isps/'
    >>> exp_list = hr.get_exp_list(savedir_base=savedir_base, 
    >>>                            filterby_list=[{'sampler':{'train':'basic'}}])
    >>> lists_of_score_lists = hr.get_score_lists(exp_list, savedir_base=savedir_base, columns=['train_loss', 'exp_id'])
    >>> print(lists_of_score_lists)
    """
    if len(exp_list) == 0:
        if verbose:
            print('exp_list is empty...')
        return

    exp_list = hu.filter_exp_list(exp_list, filterby_list, savedir_base=savedir_base, verbose=verbose)
    score_lists = []

    # aggregate results
    for exp_dict in exp_list:
        exp_id = hu.hash_dict(exp_dict)
        savedir = os.path.join(savedir_base, exp_id)

        score_list_fname = os.path.join(savedir, score_list_name)
        if not os.path.exists(score_list_fname):
            if verbose:
                print('%s: missing %s' % (exp_id, score_list_name))
            continue
        
        else:
            score_lists += [hu.load_pkl(score_list_fname)]
    
    return score_lists


def get_exp_list_df(exp_list, filterby_list=None, columns=None, verbose=True):
    """Get a table showing the configurations for the given list of experiments 

    Parameters
    ----------
    exp_list : list
        A list of experiments, each defines a single set of hyper-parameters
    columns : list, optional
        a list of columns you would like to display, by default None
        
    Returns
    -------
    DataFrame
        a dataframe showing the scores obtained by the experiments

    Example
    -------
    >>> from haven import haven_results as hr
    >>> savedir_base='../results/isps/'
    >>> exp_list = hr.get_exp_list(savedir_base=savedir_base, 
    >>>                            filterby_list=[{'sampler':{'train':'basic'}}])
    >>> df = hr.get_exp_list_df(exp_list, columns=['train_loss', 'exp_id'])
    >>> print(df)
    """
    if len(exp_list) == 0:
        if verbose:
            print('exp_list is empty...')
        return

    exp_list = hu.filter_exp_list(exp_list, filterby_list, verbose=verbose)

    # aggregate results
    result_list = []
    for exp_dict in exp_list:
        result_dict = {}

        exp_id = hu.hash_dict(exp_dict)
        result_dict["exp_id"] = exp_id

        for k in exp_dict:
            result_dict[k] = exp_dict[k]

        result_list += [result_dict]

    df = pd.DataFrame(result_list)

    if columns:
        df = df[[c for c in columns if c in df.columns]]

    return df



try:
    import ast
    from ipywidgets import Button, HBox, VBox
    from ipywidgets import widgets

    from IPython.display import display
    from IPython.core.display import Javascript, display, HTML
    from IPython.display import FileLink, FileLinks
    from ipywidgets.widgets.interaction import show_inline_matplotlib_plots
except:
    print('widgets not available...')

def autofix():
    pass

WIDTH = '200px'
class Display:
    def __init__(self, heading, options, value, db_vars, var):
        h = widgets.Label(value="Select Rows:", 
                        layout=widgets.Layout(width=WIDTH),)
        sm = widgets.SelectMultiple(options=db.rm.exp_params,
                                                    value=list(db.vars.get('latex_rows')))
        return 
    def update():
        pass


class Text:
    def __init__(self, header, default, type, db_vars, var):
        org_value = db_vars.get(var)

        if org_value is None or org_value == '':
            value = default
        else:
            value = str(org_value)
        self.type = type
        self.header = widgets.Label(value=header, 
                          layout=widgets.Layout(width=WIDTH),)
        self.text = widgets.Text(
                    value=value,
                    disabled=False,
                    layout=widgets.Layout(width=WIDTH),)
        self.db_vars = db_vars
        self.var = var
        
    def get_widget(self):
        return widgets.VBox([self.header, self.text])

    def update(self):
        self.db_vars[self.var] = str(self.text.value)
        value = self.db_vars[self.var]
        if self.type == 'tuple':
            w, h = value.strip('(').strip(')').split(',')
            value = (int(w), int(h))
        if self.type == 'int':
            value = int(value)
        if self.type == 'str':
            value = str(value)

        return value

class SelectMultiple:
    def __init__(self, header, options, db_vars, var, select_all=False):
        org_value = db_vars.get(var)

        if org_value is None:
            if len(options) > 0:
                if select_all:
                    value = options
                else:
                    value = [options[0]]
            else:
                value = []
        else:
            value = [v for v in org_value if v in options]

        self.header = widgets.Label(value=header, 
                          layout=widgets.Layout(width=WIDTH),)
        self.select_multiple = widgets.SelectMultiple(options=options, value=value, 
                          layout=widgets.Layout(width=WIDTH))
        self.db_vars = db_vars
        self.var = var
        
    def get_widget(self):
        return widgets.VBox([self.header, self.select_multiple])

    def update(self):
        self.db_vars[self.var] = list(self.select_multiple.value)
        return self.db_vars[self.var]

class Dropdown:
    def __init__(self, header, options, db_vars, var):
        org_value = db_vars.get(var)

        if org_value is None or org_value not in options:
            if len(options) > 0:
                value = options[0]
            else:
                value = None
        else:
            value = org_value

        self.header = widgets.Label(value=header, 
                          layout=widgets.Layout(width=WIDTH),)
        self.dropdown = widgets.Dropdown(
                            options=options,
                            value=value,
                            disabled=False,
                        layout=widgets.Layout(width=WIDTH),)
        self.db_vars = db_vars
        self.var = var
        
    def get_widget(self):
        return widgets.VBox([self.header, self.dropdown])

    def update(self):
        self.db_vars[self.var] = self.dropdown.value
        value = self.db_vars[self.var]
        if value == 'None':
            return None
        return value
        


def create_button():
    pass


def create_download():
    pass

def create_text_box():
    pass



import sys, os, pprint

path = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
sys.path.insert(0, path)

from haven.haven_jobs import slurm_manager as sm

if __name__ == "__main__":
  job_id = sm.submit_job()
  job = sm.get_job()
  status = sm.kill_job()



import sys, os, pprint

path = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
sys.path.insert(0, path)

import wandb
from haven import haven_wizard as hw


def filter_exp_list(exp_list, filterby_list, savedir_base=None, verbose=True,
                    score_list_name='score_list.pkl', return_style_list=False):
    """[summary]
    
    Parameters
    ----------
    exp_list : [type]
        A list of experiments, each defines a single set of hyper-parameters
    filterby_list : [type]
        [description]
    
    Returns
    -------
    [type]
        [description]
    """
    if filterby_list is None or filterby_list == '' or len(filterby_list) == 0:
        if return_style_list:
            return exp_list, [{}]*len(exp_list)
        else:
            return exp_list

    style_list = []
    filterby_list_list = hu.as_double_list(filterby_list)
    # filterby_list = filterby_list_list
    
    for filterby_list in filterby_list_list:
        exp_list_new = []

        # those with meta
        filterby_list_no_best = []
        for filterby_dict in filterby_list:
            meta_dict = {}
            if isinstance(filterby_dict, tuple):
                fd, meta_dict = filterby_dict
            
            if meta_dict.get('best'):
                assert savedir_base is not None
                el = filter_exp_list(exp_list, filterby_list=fd, verbose=verbose)
                best_dict = meta_dict.get('best')
                exp_dict = get_best_exp_dict(el, savedir_base, 
                                metric=best_dict['metric'],
                                metric_agg=best_dict['metric_agg'], 
                                filterby_list=None, 
                                avg_across=best_dict.get('avg_across'),
                                return_scores=False, 
                                verbose=verbose,
                                score_list_name=score_list_name)

                exp_list_new += [exp_dict]
                style_list += [meta_dict.get('style', {})]
            else:
                filterby_list_no_best += [filterby_dict] 

        
        # ignore metas here meta
        for exp_dict in exp_list:
            select_flag = False
            
            for fd in filterby_list_no_best:
                if isinstance(fd, tuple):
                    filterby_dict, meta_dict = fd
                    style_dict = meta_dict.get('style', {})
                else:
                    filterby_dict = fd
                    style_dict = {}

                filterby_dict = copy.deepcopy(filterby_dict)
               
                keys = filterby_dict.keys()
                for k in keys:
                    if '.' in k:
                        v = filterby_dict[k]
                        k_list = k.split('.')
                        nk = len(k_list)

                        dict_tree = dict()
                        t = dict_tree

                        for i in range(nk):
                            ki = k_list[i]
                            if i == (nk - 1):
                                t = t.setdefault(ki, v)
                            else:
                                t = t.setdefault(ki, {})

                        filterby_dict = dict_tree

                assert (isinstance(filterby_dict, dict), 
                                 ('filterby_dict: %s is not a dict' % str(filterby_dict)))

                if hu.is_subset(filterby_dict, exp_dict):
                    select_flag = True
                    break

            if select_flag:
                exp_list_new += [exp_dict]
                style_list += [style_dict]

        exp_list = exp_list_new
        

    if verbose:
        print('Filtered: %d/%d experiments gathered...' % (len(exp_list_new), len(exp_list)))
    # hu.check_duplicates(exp_list_new)
    exp_list_new = hu.ignore_duplicates(exp_list_new)
    
    if return_style_list:
        return exp_list_new, style_list

    return exp_list_new
  
if __name__ == "__main__":
  exp_list = [{'dataset':'mnist', 'model':{'name':'lenet', 'lr':1e-3}},
              {'dataset':'mnist', 'model':{'name':'kale', 'lr':1e-5}}
              ]
   
  # this works
  filterby_list = [{'model':{lr':1e-5, 'name':'kale'}}]
  
  # this might not work
  filterby_list = [{'model.lr':1e-5, 'model.name':'kale'}]
  
  # https://github.com/haven-ai/haven-ai/blob/master/haven/haven_utils/exp_utils.py#L361
  exp_list_filtered = filter_exp_list(exp_list, filterby_list)
  
  print(exp_list_filtered)



def download(url, path=None, cached=True):
    import gdown
    # kitti_url = 'https://drive.google.com/uc?id=1QHvE8oHlHqXB97RHlulLuWCmleE0wZ8B'
    # kitti_out = 'kitti.tar.bz2'
    gdown.cached_download(url, path=path, quiet=False, proxy=False, postprocess=gdown.extractall)
    # tar -xvjf kitti.tar.bz2




from .. import haven_utils as hu
from .. import haven_chk as hc
import os

import time
import copy
import pandas as pd
import numpy as np 
import getpass
import pprint
import requests

try:
    import eai_toolkit_client
    from eai_toolkit_client.rest import ApiException
except:
    pass

def to_dict(job):
    return {'id':job.id, 'state':job.state, 'runs':job.runs, 'alive':job.alive}
# Api
# ==============
def get_api(**kwargs):
    # Get Borgy API
    jobs_url = 'https://console.elementai.com'
    config = eai_toolkit_client.Configuration()
    config.host = jobs_url

    api_client = eai_toolkit_client.ApiClient(config)

    if kwargs.get('token') is None:
        try:
            token_url = 'https://internal.console.elementai.com/v1/token'
            r = requests.get(token_url)
            r.raise_for_status()
            token = r.text
        except requests.exceptions.HTTPError as errh:
            # Perhaps do something for each error
            token = hu.subprocess_call('eai login token -H').split(' ')[-1].replace('\n', '')
            
        except requests.exceptions.ConnectionError as errc:
            raise SystemExit(errc)
        except requests.exceptions.Timeout as errt:
            raise SystemExit(errt)
        except requests.exceptions.RequestException as err:
            raise SystemExit(err)

    api_client.set_default_header('Authorization', 'Bearer {}'.format(token))

    # create an instance of the API class
    api = eai_toolkit_client.JobApi(api_client)

    return api 

# Job submission
# ==============
def submit_job_v2(api, account_id, command, job_config, savedir):
    workdir = os.path.join(savedir, 'code')
    return submit_job(api, account_id, command, job_config, workdir, savedir_logs=savedir)

def submit_job(api, account_id, command, job_config, workdir, savedir_logs=None):
    job_spec = get_job_spec(job_config, command, savedir_logs, workdir=workdir)
    job = api.v1_account_job_post(account_id=account_id, human=1, job_spec=job_spec)
    job_id = job.id
    return job_id

def get_job_spec(job_config, command, savedir_logs, workdir):
    _job_config = copy.deepcopy(job_config)
    _job_config['workdir'] = workdir
    
    if savedir_logs is not None:
        path_log = os.path.join(savedir_logs, "logs.txt")
        path_err = os.path.join(savedir_logs, "err.txt")
        command_with_logs = '%s 1>%s 2>%s' % (command, path_log, path_err)
    else:
        command_with_logs = command

    _job_config['command'] = ['/bin/bash', '-c', command_with_logs]
    _job_config['resources'] = eai_toolkit_client.JobSpecResources(**_job_config['resources'])
    job_spec = eai_toolkit_client.JobSpec(**_job_config)

    # Return the Job command in Byte format
    return job_spec


# Job status
# ===========
def get_jobs_dict(api, job_id_list, query_size=20):
    # get jobs
    "id__in=64c29dc7-b030-4cb0-8c51-031db029b276,52329dc7-b030-4cb0-8c51-031db029b276"

    jobs = []
    for i in range(0, len(job_id_list), query_size):
        job_id_string = "id__in="
        for job_id in  job_id_list[i:i + query_size]:
            job_id_string += "%s," % job_id
        job_id_string = job_id_string[:-1]
        jobs += api.v1_cluster_job_get(q=job_id_string).items

    jobs_dict = {job.id: to_dict(job) for job in jobs}

    return jobs_dict

def get_job(api, job_id):
    """Get job information."""
    try:
        return to_dict(api.v1_job_get_by_id(job_id))
    except ApiException as e:
        raise ValueError("job id %s not found." % job_id)

def get_jobs(api, account_id):
    # account_id = hu.subprocess_call('eai account get').split('\n')[-2].split(' ')[0]
    return [to_dict(j) for j in api.v1_account_job_get(account_id=account_id,
            limit=1000, 
            order='-created',
            q="alive_recently=True").items]
            

    # return api.v1_me_job_get(limit=1000, 
    #         order='-created',
    #         q="alive_recently=True").items
           

# Job kill
# ===========
def kill_job(api, job_id):
    """Kill a job job until it is dead."""
    job = get_job(api, job_id)

    if not job['alive']:
        print('%s is already dead' % job_id)
    else:
        # toolkit
        api.v1_job_delete_by_id(job_id)
        print('%s CANCELLING...' % job_id)
        job = get_job(api, job_id)
        while job['state'] == "CANCELLING":
            time.sleep(2.0)
            job = get_job(api, job_id)

        print('%s now is dead.' % job_id)



import pylab as plt
import matplotlib
import numpy as np

import pylab as plt
import matplotlib
from itertools import product
import numpy as np

import pandas as pd
import os
from sklearn.metrics.pairwise import pairwise_distances
from matplotlib.ticker import ScalarFormatter, FuncFormatter
matplotlib.style.use('bmh')

markers = [("-", "o"), ("-", "p"), ("-", "D"), ("-", "^"), ("-", "s"),
           ("-", "8"), ("-", "o"), ("-", "o"), ("-", "o"), ("-", "o"),
           ("-", "o"), ("-", "o")]

colors = ['#741111', "#000000", '#3a49ba', '#7634c9',
          "#4C9950", "#CC29A3", '#ba3a3a', "#0f7265",
          "#7A7841", "#00C5CD", "#6e26d9"]

bright_colors = ["#00C5CD"]


def myticks(x, pos):

    if x == 0:
        return "$0$"

    exponent = int(np.log10(x))
    coeff = x/10**exponent

    # return r"{:0.1f}e{:0d}".format(coeff,exponent)

    return r"${:0.1f} \times 10^{{ {:2d} }}$".format(coeff, exponent)


def myticks_new(x, pos, exponent=1e5):

    if x == 0:
        return "$0$"

    exponent = int(np.log10(x))
    coeff = x/10**exponent

    return r"${:0s}$".format(coeff/exponent)

    # return r"${:0.1f} \times 10^{{ {:2d} }}$".format(coeff,exponent)


class FixedOrderFormatter(ScalarFormatter):
    """Formats axis ticks using scientific notation with a constant order of 
    magnitude"""

    def __init__(self, order_of_mag=0, useOffset=True, useMathText=False):
        self._order_of_mag = order_of_mag
        ScalarFormatter.__init__(self, useOffset=useOffset,
                                 useMathText=useMathText)

    def _set_orderOfMagnitude(self, range):
        """Over-riding this to avoid having orderOfMagnitude reset elsewhere"""
        self.orderOfMagnitude = self._order_of_mag


class PrettyPlot:
    def __init__(self, title=None, ylabel=None, xlabel=None,
                 fontsize=14, linewidth=2.5, markersize=12,
                 axFontSize=18, figsize=(13, 10), 
                 legend_type="line", yscale="log", subplots=(1, 1),
                 shareRowLabel=True, axTickSize=14, legend_size=10,
                 box_linewidth=1):
        self.box_linewidth = box_linewidth
        self.legend_size = legend_size
        self.axTickSize = int(axTickSize)
        self.fontsize = int(fontsize)
        self.shareRowLabel = shareRowLabel
        self.lim_set = False
        self.ylim = None
        self.legend_type = legend_type
        self.yscale = yscale
        self.linewidth = int(linewidth)
        self.markersize = int(markersize)
        self.axFontSize = int(axFontSize)

        if self.yscale == "log":
            plt.yscale("log")
        # ax.set_yscale('logit')
        self.labels = []
        self.y_list = []
        self.x_list = []
        self.converged = []
        fig = plt.figure(figsize=figsize)

        if title is not None:
            fig.suptitle(title, fontsize=self.axFontSize)
        self.fig = fig

        subplots = list(subplots)
        self.nrows = subplots[0]
        self.ncols = subplots[1]
        self.pIndex = 1

        self.axList = []

    def add_yxList(self, y_vals, x_vals, label, converged=False):
        if isinstance(y_vals, list):
            y_vals = np.array(y_vals)
        if isinstance(x_vals, list):
            x_vals = np.array(x_vals)

        self.y_list += [y_vals]
        self.x_list += [x_vals]

        self.labels += [label]

        self.converged += [converged]

    def show(self):
        plt.show()

    def save(self, path, iformat="png"):
        create_dirs(path)
        fname = path + ".%s" % iformat
        self.fig.savefig(fname, bbox_inches='tight')
        print(("Figure saved in %s" % (fname)))

    def plot_DataFrame(self, results):
        n_points, n_labels = results.shape

        x_vals = np.arange(n_points)
        labels = results.columns
        y_array = np.array(results)
        y_list = []
        x_list = []
        for j in range(n_labels):
            x_list += [x_vals]
            y_list += [y_array[:, j]]

        self.plot(y_list, x_list, labels)

    def set_lim(self, ylim, xlim):
        self.lim_set = True
        self.ylim = ylim

        self.ax.set_ylim(ylim)
        self.ax.set_xlim(xlim)

    def set_tickSize(self, labelsize=8):
        [tick.label.set_fontsize(labelsize)
         for tick in self.ax.yaxis.get_major_ticks()]
        [tick.label.set_fontsize(labelsize)
         for tick in self.ax.xaxis.get_major_ticks()]

    def set_title(self, title):
        self.fig.suptitle(title, fontsize=self.axFontSize, y=1.08)

    def plot(self, y_list=None, x_list=None, labels=None, ax=None,
             ylabel="", xlabel="", yscale=False):
        fig = self.fig

        if y_list == None and x_list == None:
            y_list = self.y_list
            x_list = self.x_list

        if yscale == "log":
            # Makse sure everything is non-negative
            # for yi in y_list:
            #     assert np.all(yi >= 0)

            # Set zeros to eps
            for i in range(len(y_list)):
                y_list[i] = np.maximum(y_list[i], np.finfo(float).eps)

            # Set zeros to eps
            for i in range(len(y_list)):

                opt_ind = np.where(y_list[i] == np.finfo(float).eps)[0]
                if opt_ind.size > 0:
                    opt_ind = opt_ind[0]

                    y_list[i] = y_list[i][:opt_ind+1]
                    x_list[i] = x_list[i][:opt_ind+1]

        n_labels = len(y_list)

        if ax is None:
            ax = self.fig.add_subplot(self.nrows,
                                      self.ncols, self.pIndex)

        ax.set_facecolor('white')
        ax.set_yscale("log", nonposy='clip')
        if labels is None and self.labels is None:
            labels = list(map(str, np.arange(n_labels)))
        elif labels is None:
            labels = self.labels

        ref_points = []
        for i in range(len(self.converged)):
            if self.converged[i] is not None:

                ref_points += [[self.converged[i]["X"],
                                self.converged[i]["Y"]]]

        label_positions, label_indices = get_labelPositions(y_list,
                                                            x_list,
                                                            self.ylim,
                                                            labels=labels,
                                                            ref_points=np.array(ref_points))

        ls_markers = markers

        if not self.lim_set:
            y_min, y_max = get_min_max(y_list)
            x_min, x_max = get_min_max(x_list)
            #y_min = max(y_min, 1e-8)
            ax.set_ylim([y_min, y_max])
            ax.set_xlim([x_min, x_max])

        for i in range(n_labels):
            color = colors[i]
            ls, marker = ls_markers[i]

            y_vals = y_list[i]
            x_vals = x_list[i]

            n_points = len(y_vals)

            label = labels[i]

            markerFreq = n_points / (int(np.log(n_points)) + 1)

            # SCATTER PLOT OPTIMAL
            # ind_opt = np.where(y_vals == np.finfo(float).eps)[0]

            # if ind_opt.size > 0:
            #     x_opt = x_vals[np.where(y_vals == np.finfo(float).eps)[0][0]]
            #     y_opt = np.finfo(float).eps

            if self.converged[i] is not None:
                ax.scatter(self.converged[i]["X"],
                           self.converged[i]["Y"], s=300, marker="*",  color=color, clip_on=False, zorder=100)
            ##
            line, = ax.plot(x_vals, y_vals, markevery=int(markerFreq),
                            markersize=int(self.markersize), color=color,
                            lw=self.linewidth, alpha=1.0,
                            label=label, ls=ls, marker=marker)


            if self.legend_type == "line":
                x_point, y_point = label_positions[i]
                angle = get_label_angle(
                    x_vals, y_vals, label_indices[i], ax, color='0.5', size=12)

                box = dict(facecolor="white",
                           edgecolor=color, linestyle=ls,
                           # hatch=marker,
                           linewidth=int(2), boxstyle="round")

                ax.text(x_point, y_point, label, va='center', ha='center',
                        rotation=angle,
                        color=color,
                        bbox=box,
                        fontsize=self.legend_size)

            else:
                plt.legend(loc="best")

        if self.shareRowLabel and (((self.pIndex-1) % (self.ncols)) == 0):
            ax.set_ylabel(ylabel, fontsize=self.axFontSize)

        if not self.shareRowLabel:
            ax.set_ylabel(ylabel, fontsize=self.axFontSize*1.1)

        ax.set_xlabel(xlabel, fontsize=self.axFontSize*1.1)

        ax.tick_params(labelsize=self.axTickSize*1.3)
        ax.tick_params(axis='y', labelsize=int(self.axTickSize*1.5))
        self.y_list = []
        self.x_list = []
        self.labels = []
        self.converged = []

        self.pIndex += 1
        self.axList += [ax]

        ax.minorticks_off()
        vals = np.logspace(np.log10(y_min), np.log10(y_max), 5)
        ax.set_yticks(vals)

        ax.yaxis.set_major_formatter(FuncFormatter(myticks))

        return fig, ax


#########
# HELPERS
#########
def get_overlapPercentage(index, y_list):
    n_points = y_list[0].size
    for i in range(index + 1):
        n_points = min(n_points, y_list[i].size)

    y_vector = y_list[index][:n_points, np.newaxis]

    prev_lines = np.zeros((n_points, index))

    for i in range(index):
        prev_lines[:, i] = y_list[i][:n_points]

        prev_lines /= (np.linalg.norm(prev_lines, axis=0) + 1e-10)

    y_norm = y_vector / np.linalg.norm(y_vector, axis=0)

    diff = np.abs((prev_lines - y_norm)).min(axis=1)

    n_overlap = np.sum(diff < 1e-6)

    percentage = n_overlap / float(n_points)

    return percentage


def create_dirs(fname):
    if "/" not in fname:
        return

    if not os.path.exists(os.path.dirname(fname)):
        try:
            os.makedirs(os.path.dirname(fname))
        except OSError:
            pass


def normalize(xy_points, ref_points, y_min, y_max, x_min, x_max):

    xy_points[:, 1] = np.log(np.maximum(1e-15, xy_points[:, 1])) / np.log(10)
    ref_points[:, 1] = np.log(np.maximum(ref_points[:, 1], 1e-15)) / np.log(10)

    y_min = np.log(y_min)/np.log(10)
    y_max = np.log(y_max)/np.log(10)

    xy_normed = xy_points - np.array([x_min, y_min])
    xy_normed /= np.array([x_max - x_min, y_max - y_min])

    ref_normed = ref_points - np.array([x_min, y_min])

    ref_normed /= np.array([x_max - x_min, y_max - y_min])
    return xy_normed, ref_normed

# LABEL POSITIONS


def get_labelPositions(y_list, x_list, ylim=None, labels=None, ref_points=None):
    if ref_points is None:

        ref_points = []
    """Get label positions greedily"""

    n_labels = len(y_list)

    # GET BORDER POINTS
    x_min, x_max = get_min_max(x_list)
    if ylim is not None:
        y_min, y_max = ylim
    else:
        y_min, y_max = get_min_max(y_list)

    xdiff = (x_max - x_min)
    ydiff = (y_max - y_min)

    # Border points
    bp1 = np.array(list(product([x_min, x_max, xdiff * 0.5],
                                [y_min, y_max, ydiff * 0.5])))[:-1]
    bp1 = np.array(list(product([x_max],
                                [y_max])))[:-1]

    bp1 = np.array(list(product([8],
                                [0])))

    addedPoints = []

    for yPoint in np.linspace(y_min, y_max, 6):
        addedPoints += [(x_min, yPoint)]
        addedPoints += [(x_max, yPoint)]

    sPoints = [(xx[0], yy[0]) for xx, yy in zip(x_list, y_list)]
    ePoints = [(xx[-1], yy[-1]) for xx, yy in zip(x_list, y_list)]
    bp2 = np.array(addedPoints + sPoints + ePoints)
    if len(ref_points) == 0:
        border_points = np.vstack([bp1, bp2])
    else:
        border_points = np.vstack([bp1, bp2, ref_points])

    n_border = border_points.shape[0]

    # Initialize placeholders
    ref_points = np.zeros((n_border + n_labels, 2))

    label_positions = np.zeros((n_labels, 2))
    label_indices = np.zeros(n_labels, int)

    ref_points[:n_border] = border_points

    for i in range(n_labels):
        # GET POSITIONS

        if ylim is not None:
            ind = (y_list[i] < y_max+1e-4) & (y_list[i] > y_min-1e-4)
            n_points = x_list[i][ind].size
            xy_points = np.zeros((n_points, 2))
            xy_points[:, 0] = x_list[i][ind]
            xy_points[:, 1] = y_list[i][ind]
        else:
            n_points = x_list[i].size
            xy_points = np.zeros((n_points, 2))
            xy_points[:, 0] = x_list[i]
            xy_points[:, 1] = y_list[i]

        # NORMALIZE

        xy_normed, ref_normed = normalize(xy_points.copy(),
                                          ref_points[:n_border+i].copy(), y_min, y_max, x_min, x_max)
        # GET REF POINTS

        dist = pairwise_distances(xy_normed, ref_normed,
                                  metric="l1")

        # GET MINIMUM DISTANCES
        min_dist = dist.min(axis=1)

        # GET MAXIMUM MINIMUM DISTANCE
        label_index = np.argmax(min_dist)
        label_pos = xy_points[label_index]

        ref_points[n_border + i] = label_pos
        label_positions[i] = label_pos
        label_indices[i] = label_index

    return label_positions, label_indices


def get_min_max(v_list):
    vector = v_list[0]
    v_min = np.min(vector)
    v_max = np.max(vector)

    for i in range(1, len(v_list)):
        vector = v_list[i]
        v_min = min(np.min(vector), v_min)
        v_max = max(np.max(vector), v_max)

    return v_min, v_max


def get_label_angle(xdata, ydata, index, ax, color='0.5', size=12, window=3):
    n_points = xdata.size

    x1 = xdata[index]
    y1 = ydata[index]

    #ax = line.get_axes()

    sp1 = ax.transData.transform_point((x1, y1))

    slope_degrees = 0.
    count = 0.

    for i in range(index+1, min(index+window, n_points)):
        y2 = ydata[i]
        x2 = xdata[i]

        sp2 = ax.transData.transform_point((x2, y2))

        rise = (sp2[1] - sp1[1])
        run = (sp2[0] - sp1[0])

        slope_degrees += np.degrees(np.arctan2(rise, run))
        count += 1.

    for i in range(index-1, max(index-window, 0), -1):
        y2 = ydata[i]
        x2 = xdata[i]

        sp2 = ax.transData.transform_point((x2, y2))

        rise = - (sp2[1] - sp1[1])
        run = -(sp2[0] - sp1[0])

        slope_degrees += np.degrees(np.arctan2(rise, run))
        count += 1.

    slope_degrees /= count

    return slope_degrees


def box_color(edgecolor, linestyle, marker):
    """Creates box shape"""
    return dict(facecolor="white",
                edgecolor=edgecolor, linestyle=linestyle,
                # hatch=marker,
                linewidth=2, boxstyle="round")



from .. import haven_utils as hu
from .. import haven_chk as hc
import os

import time
import copy
import pandas as pd
import numpy as np
import getpass
import pprint
import requests
import pandas as pd

# Job submission
# ==============
def submit_job(api, account_id, command, job_config, workdir, savedir_logs=None):
    # read slurm setting
    lines = "#! /bin/bash \n"
    lines += "#SBATCH --account=%s \n" % account_id
    for key in list(job_config.keys()):
        lines += "#SBATCH --%s=%s \n" % (key, job_config[key])
    path_log = os.path.join(savedir_logs, "logs.txt")
    lines += "#SBATCH --output=%s \n" % path_log
    path_err = os.path.join(savedir_logs, "err.txt")
    lines += "#SBATCH --error=%s \n" % path_err
    path_code = os.path.join(savedir_logs, "code")
    lines += "#SBATCH --chdir=%s \n" % path_code

    lines += command

    file_name = os.path.join(savedir_logs, "bash.sh")
    hu.save_txt(file_name, lines)

    # launch the exp
    submit_command = "sbatch %s" % file_name
    while True:
        try:
            job_id = hu.subprocess_call(submit_command).split()[-1]
        except Exception as e:
            if "Socket timed out" in str(e):
                print("slurm time out and retry now")
                time.sleep(1)
                continue
        break

    # delete the bash.sh
    os.remove(file_name)

    return job_id
    

# Job status
# ===========
def get_job(api, job_id):
    """Get job information."""
    job_info = get_jobs_dict(None, [job_id])[job_id]
    job_info["job_id"] = job_id

    return job_info


def get_jobs(api, account_id):
    ''' get all jobs launched by the current user'''
    job_list = ""
    command = "squeue --user=%s --format=\"%%.18i %%.8T\"" % getpass.getuser()
    while True:
        try:
            job_list = hu.subprocess_call(command)
        except Exception as e:
            if "Socket timed out" in str(e):
                print("squeue time out and retry now")
                time.sleep(1)
                continue
        break

    result = [{"job_id": j.split()[0], "state": j.split()[1]} for j in job_list.split('\n')[1:-1]]
    return result


def get_jobs_dict(api, job_id_list, query_size=20):
    if len(job_id_list) == 0:
        return {}

    jobs_dict = {}

    command = "sacct --jobs=%s --format=jobid,cputime,state" % str(job_id_list)[1:-1].replace(" ", "")
    while True:
        try:
            job_list = hu.subprocess_call(command)
        except Exception as e:
            if "Socket timed out" in str(e):
                print("sacct time out and retry now")
                time.sleep(1)
                continue
        break

    lines = job_list.split('\n')
    header = lines[0].split()
    lines = [l.split() for l in lines[2:-1]]

    df = pd.DataFrame(data=lines, columns=header)
    df = df[~df["JobID"].str.contains(r"\.")]
    df = df.rename(mapper={"State": "state", "CPUTime": "cpuTime", "JobID": "job_id"}, axis=1)
    df = df.replace({"state": r"CANCELLED.*"}, {"state": "CANCELLED"}, regex=True)
    df.insert(loc=0, column="runs", value="")

    # use job id as key
    new_df = df.drop(labels="job_id", axis=1)
    new_df.index = df["job_id"].to_list()
    jobs_dict = new_df.to_dict(orient="index")

    return jobs_dict

# Job kill
# ===========


def kill_job(api, job_id):
    """Kill a job job until it is dead."""
    job = get_job(api, job_id)

    if job["state"] in ["CANCELLED", "COMPLETED", "FAILED", "TIMEOUT"]:
        print('%s is already dead' % job_id)
    else:
        kill_command = "scancel %s" % job_id
        while True:
            try:
                hu.subprocess_call(kill_command)
                print('%s CANCELLING...' % job_id)
            except Exception as e:
                if "Socket timed out" in str(e):
                    print("scancel time out and retry now")
                    time.sleep(1)
                    continue
            break

        # confirm cancelled
        job = get_job(api, job_id)
        while job["state"] != "CANCELLED":
            time.sleep(2)
            job = get_job(api, job_id)

        print('%s now is dead.' % job_id)




# Transformers
import collections
import numpy as np 
import torch 

from torchvision import transforms

mean = [0.485, 0.456, 0.406]
std = [0.229, 0.224, 0.225]


def get_transformer(transform, split):
    if transform == "resize_normalize":
        normalize_transform = transforms.Normalize(mean=mean, std=std)

        return transforms.Compose(
            [transforms.Resize((224,224)),
             transforms.ToTensor(),
             normalize_transform])

    if transform == "rgb_normalize":
        normalize_transform = transforms.Normalize(mean=mean, std=std)

        return transforms.Compose(
            [
             transforms.ToTensor(),
             normalize_transform])

    elif transform == "resize":
        return ComposeJoint(
            [
                [transforms.Resize((224, 224)), transforms.Resize((224, 224)), None],
                [transforms.ToTensor(), None, transforms.ToTensor()],
                [transforms.Normalize(mean=mean, std=std), None, None],
                [None, ToLong(), None]
            ])

    elif transform == "basic":
        transform = ComposeJoint(
            [
                [transforms.ToTensor(), None],
                [transforms.Normalize(mean=mean, std=std), None],
                [None, ToLong()]
            ])
        return transform

    elif transform == "basic_flip":
        transform = ComposeJoint(
            [   RandomHorizontalFlipJoint(),      
                [transforms.ToTensor(), None, transforms.ToTensor()],
                [transforms.Normalize(mean=mean, std=std), None, None],
                [None, ToLong(), None]
            ])
        return transform

# helpers
class ComposeJoint(object):
    def __init__(self, transforms):
        self.transforms = transforms

    def __call__(self, x):
        for transform in self.transforms:
            x = self._iterate_transforms(transform, x)

        return x

    def _iterate_transforms(self, transforms, x):
        """Credit @fmassa:
         https://gist.github.com/fmassa/3df79c93e82704def7879b2f77cd45de
        """
        if isinstance(transforms, collections.Iterable):
            for i, transform in enumerate(transforms):
                x[i] = self._iterate_transforms(transform, x[i])
        else:

            if transforms is not None:
                x = transforms(x)

        return x

class ToLong(object):
    def __call__(self, x):
        return torch.LongTensor(np.asarray(x))


import os
import time
import sys
import subprocess
from .. import haven_utils as hu
from .. import haven_chk as hc
import os
from textwrap import wrap
import time
import copy
import pandas as pd
import numpy as np
import getpass
import pprint

ALIVE_STATES =  ['RUNNING', 'QUEUED', 'PENDING', 'QUEUING']
COMPLETED_STATES =  ['COMPLETED', 'SUCCEEDED', 'COMPLETING']
FAILED_STATES = ["FAILED", "CANCELLED", "INTERRUPTED", "TIMEOUT"]


class JobManager:
    """Job manager."""

    def __init__(self,
                 exp_list=None,
                 savedir_base=None,
                 workdir=None,
                 job_config=None,
                 verbose=1,
                 account_id=None):
        """[summary]

        Parameters
        ----------
        exp_list : [type]
            [description]
        savedir_base : [type]
            [description]
        workdir : [type], optional
            [description], by default None
        job_config : [type], optional
            [description], by default None
        verbose : int, optional
            [description], by default 1
        """
        if account_id is None and job_config is not None and 'account_id' in job_config:
            account_id = job_config['account_id']
            del job_config['account_id']

        self.exp_list = exp_list
        self.job_config = job_config
        self.workdir = workdir
        self.verbose = verbose
        self.savedir_base = savedir_base
        self.account_id = account_id

        # define funcs
        try:
            from . import toolkit_manager as ho
            self.ho = ho
            self.api = self.ho.get_api(token=None)
            print('using toolkit')
        except:
            import slurm_manager as ho
            self.ho = ho
            self.api = None
            print('using slurm')

    
    def get_command_history(self, topk=10):
        job_list = self.get_jobs()

        count = 0
        for j in job_list:
            if hasattr(j, 'command'):
                print(count,':',j.command[2])
            if count > topk:
                break
            count += 1

    # Base functions
    # --------------
    def get_jobs(self):
        return self.ho.get_jobs(self.api, account_id=self.account_id)

    def get_jobs_dict(self, job_id_list):
        return self.ho.get_jobs_dict(self.api, job_id_list)

    def get_job(self, job_id):
        return self.ho.get_job(self.api, job_id)

    def kill_job(self, job_id):
        return self.ho.kill_job(self.api, job_id)

    def submit_job(self, command, workdir, savedir_logs=None):
        return self.ho.submit_job(api=self.api, 
                             account_id=self.account_id, 
                             command=command, 
                             job_config=self.job_config, 
                             workdir=workdir,
                             savedir_logs=savedir_logs)

    # Main functions
    # --------------
    def launch_menu(self, command=None, exp_list=None, get_logs=False, wait_seconds=3,
                    in_parallel=True):
        exp_list = exp_list or self.exp_list
        summary_list = self.get_summary_list(get_logs=False, exp_list=exp_list)
        summary_dict = hu.group_list(summary_list, key='job_state', return_count=True)

        print("\nTotal Experiments:", len(exp_list))
        print("Experiment Status:", summary_dict)
        prompt = ("\nMenu:\n"
                  "  0)'ipdb' run ipdb for an interactive session; or\n"
                  "  1)'reset' to reset the experiments; or\n"
                  "  2)'run' to run the remaining experiments and retry the failed ones; or\n"
                  "  3)'status' to view the job status; or\n"
                  "  4)'kill' to kill the jobs.\n"
                  "Type option: "
                  )
        
        option = input(prompt)

        option_list = ['reset', 'run', 'status', 'logs', 'kill']
        if option not in option_list:
            raise ValueError(
                'Prompt input has to be one of these choices %s' % option_list)

        if option == 'ipdb':
            import ipdb; ipdb.set_trace()
            print('Example:\nsummary_dict = self.get_summary(get_logs=True, exp_list=exp_list)')

        elif option == 'status':
            # view experiments
            self.print_job_status(exp_list=exp_list)
            return

        elif option == 'reset':
            self.verbose = False
            for e_list in chunk_list(exp_list, n=100):
                self.launch_exp_list(command=command, exp_list=e_list, reset=1, in_parallel=in_parallel)

        elif option == 'run':
            self.verbose = False
            tmp_list = [s_dict['exp_dict'] for s_dict in summary_list 
                            if s_dict['job_state'] 
                            not in ALIVE_STATES + COMPLETED_STATES]

            print('Selected %d/%d exps' % (len(tmp_list), len(exp_list)))
            exp_list = tmp_list
            if len(exp_list) == 0:
                print('All experiments have ran.')
                return

            for e_list in chunk_list(exp_list, n=100):
                self.launch_exp_list(command=command, exp_list=e_list, reset=0, in_parallel=in_parallel)

        elif option == 'kill':
            self.verbose = False
            tmp_list = [s_dict['exp_dict'] for s_dict in
                         summary_list if s_dict['job_state'] in
                          ALIVE_STATES]
            self.kill_jobs(exp_list=tmp_list)

        # view experiments
        # print("Checking job status...")
        time.sleep(wait_seconds)
        self.print_job_status(exp_list=exp_list)

    def print_job_status(self, exp_list):
        summary_list = self.get_summary_list(get_logs=False, exp_list=exp_list)
        summary_dict = hu.group_list(summary_list, key='job_state', return_count=False)
        
        for k in summary_dict.keys():
            n_jobs = len(summary_dict[k])
            if n_jobs:
                # print('\nExperiments %s: %d' % (k, n_jobs))
                pass
                # print(pd.DataFrame(summary_dict[k]).head())

        summary_dict = hu.group_list(summary_list, key='job_state', return_count=True)
        print(summary_dict)

    def launch_exp_list(self, command,  exp_list=None, savedir_base=None, reset=0, in_parallel=True):
        exp_list = exp_list or self.exp_list
        assert '<exp_id>' in command

        submit_dict = {}

        if in_parallel:
            pr = hu.Parallel()

            for exp_dict in exp_list:
                exp_id = hu.hash_dict(exp_dict)

                savedir_base = savedir_base or self.savedir_base
                savedir = os.path.join(savedir_base, hu.hash_dict(exp_dict))

                com = command.replace('<exp_id>', exp_id)
                pr.add(self.launch_or_ignore_exp_dict, exp_dict, com, reset, savedir, submit_dict)

            pr.run()
            pr.close()

        else:
            for exp_dict in exp_list:
                exp_id = hu.hash_dict(exp_dict)

                savedir_base = savedir_base or self.savedir_base
                savedir = os.path.join(savedir_base, hu.hash_dict(exp_dict))

                com = command.replace('<exp_id>', exp_id)
                self.launch_or_ignore_exp_dict(exp_dict, com, reset, savedir, submit_dict)

        if len(submit_dict) == 0:
            raise ValueError('The threads have an error, most likely a permission error (see above)')
        
        for i, (k,v) in enumerate(submit_dict.items()):
            print('***')
            print('Exp %d/%d - %s' % (i+1, len(submit_dict), v['message']))
            print('exp_id: %s' % hu.hash_dict(v['exp_dict']))
            pprint.pprint(v['exp_dict'])
            print()

        print("%d experiments submitted." % len(exp_list))                                      
        if len(submit_dict) > 0:
            assert len(submit_dict) == len(exp_list), 'considered exps does not match expected exps'
        return submit_dict

    def kill_jobs(self, exp_list=None):
        exp_list = exp_list or self.exp_list
        hu.check_duplicates(exp_list)

        pr = hu.Parallel()
        submit_dict = {}

        for exp_dict in exp_list:
            exp_id = hu.hash_dict(exp_dict)
            savedir = os.path.join(self.savedir_base, exp_id)
            fname = get_job_fname(savedir)

            if os.path.exists(fname):
                job_id = hu.load_json(fname)['job_id']
                pr.add(self.kill_job, job_id)
                submit_dict[exp_id] = 'KILLED'
            else:
                submit_dict[exp_id] = 'NoN-Existent'

        pr.run()
        pr.close()
        pprint.pprint(submit_dict)
        print("%d/%d experiments killed." % (len([s for s in submit_dict.values() if 'KILLED' in s]),
                                             len(submit_dict)))
        return submit_dict

    def launch_or_ignore_exp_dict(self, exp_dict, command, reset, savedir, submit_dict={}):
        """launch or ignore job.

        It checks if the experiment exist and manages the special casses, e.g.,
        new experiment, reset, failed, job is already running, completed
        """
        # Define paths
        fname = get_job_fname(savedir)

        if not os.path.exists(fname):
            # Check if the job already exists
            job_dict = self.launch_exp_dict(exp_dict, savedir, command, job=None)
            job_id = job_dict['job_id']
            message = "SUBMITTED: Launching"

        elif reset:
            # Check if the job already exists
            job_id = hu.load_json(fname).get("job_id")
            self.kill_job(job_id)
            hc.delete_and_backup_experiment(savedir)

            job_dict = self.launch_exp_dict(exp_dict, savedir, command, job=None)
            job_id = job_dict['job_id']
            message = "SUBMITTED: Resetting"

        else:
            job_id = hu.load_json(fname).get("job_id")
            job = self.get_job( job_id)

            if job['state'] in ALIVE_STATES + COMPLETED_STATES:
                # If the job is alive, do nothing
                message = 'IGNORED: Job %s' % job['state']

            elif job['state'] in FAILED_STATES:
                message = "SUBMITTED: Retrying %s Job" % job['state']
                job_dict = self.launch_exp_dict(exp_dict, savedir, command, job=job)
                job_id = job_dict['job_id']
            # This shouldn't happen
            else:
                raise ValueError('wtf')

        submit_dict[job_id] = {'exp_dict':exp_dict, 'message':message}

    def launch_exp_dict(self, exp_dict, savedir, command, job=None):
        """Submit a job job and save job dict and exp_dict."""
        # Check for duplicates
        # if job is not None:
            # assert self._assert_no_duplicates(job)

        fname_exp_dict = os.path.join(savedir, "exp_dict.json")
        hu.save_json(fname_exp_dict, exp_dict)
        exp_id = hu.hash_dict(exp_dict)
        assert(hu.hash_dict(hu.load_json(fname_exp_dict)) == exp_id)


        # Define paths
        workdir_job = os.path.join(savedir, "code")

        # Copy the experiment code into the experiment folder
        print(f'Copying code for experiment {exp_id}')
        hu.copy_code(self.workdir + "/", workdir_job, verbose=0)

        # Run  command
        job_id = self.submit_job(command, workdir_job, savedir_logs=savedir)
        print(f'Job submitted for experiment {exp_id} with job id {job_id}')
        
        # Verbose
        if self.verbose:
            print("Job_id: %s command: %s" % (job_id, command))

        job_dict = {"job_id": job_id,
                    "command": command}

        hu.save_json(get_job_fname(savedir), job_dict)

        return job_dict


    def get_summary_list(self, failed_only=False, columns=None, max_lines=10, wrap_size=8,
                               add_prefix=False, get_logs=True, exp_list=None, savedir_base=None):
        savedir_base = savedir_base or self.savedir_base
        exp_list = exp_list or self.exp_list

        # get job key
        job_id_list = []
        for exp_dict in exp_list:
            exp_id = hu.hash_dict(exp_dict)
            savedir = os.path.join(savedir_base, exp_id)
            fname = get_job_fname(savedir)

            if os.path.exists(fname):
                job_id_list += [hu.load_json(fname)["job_id"]]

        jobs_dict = self.get_jobs_dict(job_id_list)

        # get summaries
        summary_list = []
        
        for exp_dict in exp_list:
            result_dict = {}

            exp_id = hu.hash_dict(exp_dict)
            savedir = os.path.join(savedir_base, exp_id)
            job_fname = get_job_fname(savedir)

            # General info
            result_dict = {}
            result_dict['exp_dict'] = exp_dict
            result_dict["exp_id"] = exp_id
            result_dict["job_id"] = None
            result_dict["job_state"] = 'NEVER LAUNCHED'

            if os.path.exists(job_fname):
                job_dict = hu.load_json(job_fname)
                job_id = job_dict["job_id"]
                if job_id not in jobs_dict:
                    continue

                fname_exp_dict = os.path.join(savedir, "exp_dict.json")
                job = jobs_dict[job_id]
                

                if hasattr(job, 'command'):
                    command = job_dict['command']
                else:
                    command = None

                # Job info
                result_dict['started_at'] = hu.time_to_montreal(fname_exp_dict)
                result_dict["job_id"] = job_id
                result_dict["job_state"] = job["state"]
                result_dict["restarts"] = len(job["runs"])
                result_dict["command"] = command
                
                if get_logs:
                    # Logs info
                    if job["state"] == "FAILED":
                        logs_fname = os.path.join(savedir, "err.txt")
                    else:
                        logs_fname = os.path.join(savedir, "logs.txt")

                    if os.path.exists(logs_fname):
                        result_dict["logs"] = hu.read_text(logs_fname)[-max_lines:]

            summary_list += [result_dict]

        return summary_list

    def get_summary(self, **kwargs):
        """[summary]

        Returns
        -------
        [type]
            [description]
        """
        # get job ids

        # fill summary
        summary_dict = {'table': [], 
                        'status': [],
                        'logs_failed': [], 
                        'logs': []}

        # get info
        df = pd.DataFrame(summary_dict['table'])

        # if columns:
        #     df = df[[c for c in columns if (c in df.columns and c not in ['err'])]]

        if "job_state" in df:
            stats = np.vstack(
                np.unique(df['job_state'].fillna("NaN"), return_counts=True)).T
            status = ([{a: b} for (a, b) in stats])
        else:
            df['job_state'] = None

        df = hu.sort_df_columns(df)
        summary_dict['status'] = status
        summary_dict['table'] = df

        for state in ALIVE_STATES + COMPLETED_STATES + FAILED_STATES:
            summary_dict[state] = df[df['job_state'] == state]

        return summary_dict

    def _assert_no_duplicates(self, job_new=None, max_jobs=500):
        # Get the job list
        jobList = self.get_jobs()

        # Check if duplicates already exist in job
        command_dict = {}
        for job in jobList:

            if hasattr(job, 'command'):
                if job.command is None:
                    continue
                job_python_command = job.command[2]
            else:
                job_python_command = None

            if job_python_command is None:
                continue
            elif job_python_command not in command_dict:
                command_dict[job_python_command] = job
            else:
                print("Job state", job['state'], "Job command",
                      job_python_command)
                raise ValueError("Job %s is duplicated" % job_python_command)

        # Check if the new job causes duplicate
        if job_new is not None:
            if job_new.command[2] in command_dict:
                job_old_id = command_dict[job_new.command[2]].id
                raise ValueError("Job exists as %s" % job_old_id)

        return True


def get_job_fname(savedir):
    if os.path.exists(os.path.join(savedir, "borgy_dict.json")):
        # for backward compatibility
        fname = os.path.join(savedir, "borgy_dict.json")
    else:
        fname = os.path.join(savedir, 'job_dict.json')

    return fname



def run_exp_list_jobs(exp_list,
                      savedir_base,
                      workdir,
                      run_command,
                      job_config=None,
                      force_run=False,
                      wait_seconds=3,
                      account_id=None):
    """Run the experiments in the cluster.

    Parameters
    ----------
    exp_list : list
        list of experiment dictionaries
    savedir_base : str
        the directory where the experiments are saved
    workdir : str
        main directory of the code
    run_command : str
        the command to be ran in the cluster
    job_config : dict
        dictionary describing the job specifications

    Example
    -------
    Add the following code to the main file.

    >>> elif args.run_jobs:
    >>>    from haven import haven_jobs as hjb
    >>>    job_config = {'data': <data>,
    >>>                  'image': <image>,
    >>>                  'bid': '1',
    >>>                  'restartable': '1',
    >>>                  'gpu': '1',
    >>>                  'mem': '20',
    >>>                  'cpu': '2'}
    >>>    run_command = ('python trainval.py -ei <exp_id> -sb %s' %  (args.savedir_base))
    >>>    hjb.run_exp_list_jobs(exp_list, 
    >>>                          savedir_base=args.savedir_base, 
    >>>                          workdir=os.path.dirname(os.path.realpath(__file__)),
    >>>                          run_command=run_command,
    >>>                          job_config=job_config)
    """
    # let the user choose one of these options
    jm = JobManager(exp_list,
                    savedir_base,
                    run_command=run_command,
                    workdir=workdir,
                    job_config=job_config,
                    verbose=1,
                    account_id=account_id,
                    token=token)

    jm.run()

def chunk_list(my_list, n=100):
    return [my_list[x:x+n] for x in range(0, len(my_list), n)]


import os
import argparse
import pandas as pd
import pprint

from . import haven_utils as hu


def get_args():
    parser = argparse.ArgumentParser(formatter_class=make_wide(argparse.ArgumentDefaultsHelpFormatter))

    parser.add_argument('-e', '--exp_group_list', nargs="+",
                        help='Define which exp groups to run.')
    parser.add_argument('-sb', '--savedir_base', default=None,
                        help='Define the base directory where the experiments will be saved.')
    parser.add_argument('-d', '--datadir', default=None,
                        help='Define the dataset directory.')
    parser.add_argument("-r", "--reset",  default=0, type=int,
                        help='Reset or resume the experiment.')
    parser.add_argument("-ei", "--exp_id", default=None,
                        help='Run a specific experiment based on its id.')
    parser.add_argument("-j", "--run_jobs", default=0, type=int,
                        help='Run the experiments as jobs in the cluster.')
    parser.add_argument("-nw", "--num_workers", type=int, default=0,
                        help='Specify the number of workers in the dataloader.')
    parser.add_argument("-v", "--visualize_notebook", type=str, default='',
                        help='Create a jupyter file to visualize the results.')

    args, others = parser.parse_known_args()

    return args

def make_wide(formatter, w=120, h=36):
    """Return a wider HelpFormatter, if possible."""
    try:
        # https://stackoverflow.com/a/5464440
        # beware: "Only the name of this class is considered a public API."
        kwargs = {'width': w, 'max_help_position': h}
        formatter(None, **kwargs)
        return lambda prog: formatter(prog, **kwargs)
    except TypeError:
        import warnings
        warnings.warn("argparse help formatter failed, falling back.")
        return formatter

def run_wizard(func, exp_list=None, exp_groups=None, job_config=None, 
                savedir_base=None, 
               reset=None, args=None, use_threads=False,
               exp_id=None):
    if args is None:
        args = get_args()

    # Asserts
    # =======
    savedir_base = savedir_base or args.savedir_base
    reset = reset or args.reset
    exp_id = exp_id or args.exp_id
    assert savedir_base is not None

    # Collect experiments
    # ===================
    if exp_id is not None:
        # select one experiment
        savedir = os.path.join(savedir_base, exp_id)
        exp_dict = hu.load_json(os.path.join(savedir, "exp_dict.json"))

        exp_list = [exp_dict]

    elif exp_list is None:
        # select exp group
        exp_list = []
        for exp_group_name in args.exp_group_list:
            exp_list += exp_groups[exp_group_name]

    # save results folder
    if exp_id is None:
        results_fname = args.visualize_notebook
        if len(results_fname):
            if '.ipynb' not in results_fname:
                results_fname += '.ipynb'
            create_jupyter_file(fname=results_fname,
                                savedir_base=savedir_base)

    hu.check_duplicates(exp_list)
    print('\nRunning %d experiments' % len(exp_list))

    # Run experiments
    # ===============
    if not args.run_jobs:
        for exp_dict in exp_list:
            savedir = create_experiment(exp_dict, savedir_base, reset=reset,
                                        verbose=True)
            # do trainval
            func(exp_dict=exp_dict,
                 savedir=savedir,
                 args=args)
    else:
        # launch jobs
        from haven import haven_jobs as hjb
        assert job_config is not None
        assert 'account_id' in job_config
        jm = hjb.JobManager(exp_list=exp_list,
                            savedir_base=savedir_base,
                            workdir=os.getcwd(),
                            job_config=job_config,
                            )

        command = ('python trainval.py -ei <exp_id> -sb %s -d %s' %
                   (savedir_base, args.datadir))

        print(command)
        jm.launch_menu(command=command, in_parallel=use_threads)


def create_jupyter_file(fname, savedir_base):
    if not os.path.exists(fname):
        cells = [main_cell(savedir_base)]
        save_ipynb(fname, cells)
        print('> Open %s to visualize results' % fname)


def save_ipynb(fname, script_list):
    import nbformat as nbf

    nb = nbf.v4.new_notebook()
    nb['cells'] = [nbf.v4.new_code_cell(code) for code in
                   script_list]
    with open(fname, 'w') as f:
        nbf.write(nb, f)


def main_cell(savedir_base):
    script = ("""
from haven import haven_jupyter as hj
from haven import haven_results as hr
from haven import haven_utils as hu

# path to where the experiments got saved
savedir_base = '%s'
exp_list = None

# filter exps
# e.g. filterby_list =[{'dataset':'mnist'}] gets exps with mnist
filterby_list = None

# get experiments
rm = hr.ResultManager(exp_list=exp_list, 
                      savedir_base=savedir_base, 
                      filterby_list=filterby_list,
                      verbose=0,
                      exp_groups=None
                     )

# launch dashboard
# make sure you have 'widgetsnbextension' enabled; 
# otherwise see README.md in https://github.com/haven-ai/haven-ai

hj.get_dashboard(rm, vars(), wide_display=False, enable_datatables=False)
          """ % savedir_base)
    return script


def create_experiment(exp_dict, savedir_base, reset, copy_code=False, return_exp_id=False, verbose=True):
    import pprint
    from . import haven_chk as hc

    exp_id = hu.hash_dict(exp_dict)
    savedir = os.path.join(savedir_base, exp_id)

    if reset:
        hc.delete_and_backup_experiment(savedir)

    # create experiment structure
    os.makedirs(savedir, exist_ok=True)

    #-- exp_dict
    exp_dict_json_fname = os.path.join(savedir, "exp_dict.json")
    # if not os.path.exists(exp_dict_json_fname):
    hu.save_json(exp_dict_json_fname, exp_dict)

    #-- images
    os.makedirs(os.path.join(savedir, 'images'), exist_ok=True)

    if copy_code:
        src = os.getcwd() + "/"
        dst = os.path.join(savedir, 'code')
        hu.copy_code(src, dst)

    if verbose:
        print('\n******')
        print('Exp id: %s' % exp_id)
        print('\nHyperparameters:\n' + "-"*16)
        # print(pd.DataFrame([exp_dict]).to_string(index=False))
        pprint.pprint(exp_dict)

        print("\nSave directory: %s" % savedir)
        print("=" * 100)
        

    if return_exp_id:
        return savedir, exp_id

    return savedir



class Checkpointer:
    def __init__(self, savedir, return_model_state_dict=False, verbose=True):
        self.savedir = savedir
        self.verbose = verbose
        self.chk_dict = get_checkpoint(savedir, return_model_state_dict=return_model_state_dict)

    def save_checkpoint(self, score_dict, score_list=None, model_state_dict=None, images=None, 
                        images_fname=None, fname_suffix=''):
        
        save_checkpoint(savedir, score_list, model_state_dict=model_state_dict,
                    images=images, images_fname=images_fname, 
                    fname_suffix=fname_suffix, verbose=self.verbose)
    
def save_checkpoint(savedir, score_list, model_state_dict=None,
                    images=None, images_fname=None, fname_suffix='', verbose=True):
    # Report
    if verbose:
        exp_dict = hu.load_json(os.path.join(savedir, 'exp_dict.json'))
        
        print('\nExp id: %s' % hu.hash_dict(exp_dict))

        print('\nHyperparameters:\n' + "-"*16)
        # print(pd.DataFrame([exp_dict]).to_string(index=False))
        pprint.pprint(exp_dict)
        print('\nMetrics:\n' + "-"*8)
        score_df = pd.DataFrame(score_list)
        print(score_df.tail().to_string(index=False), "\n")

        
        print("Save directory: %s" % savedir)

    # save score_list
    score_list_fname = os.path.join(savedir, 'score_list%s.pkl' % fname_suffix)
    hu.save_pkl(score_list_fname, score_list)
    # if verbose:
        # print('> Saved "score_list" as %s' %
        #       os.path.split(score_list_fname)[-1])

    # save model
    if model_state_dict is not None:
        model_state_dict_fname = os.path.join(
            savedir, 'model%s.pth' % fname_suffix)
        hu.torch_save(model_state_dict_fname, model_state_dict)
        # if verbose:
            # print('> Saved "model_state_dict" as %s' %
            #       os.path.split(model_state_dict_fname)[-1])

    # save images
    images_dir = os.path.join(savedir, 'images%s' % fname_suffix)
    if images is not None:
        for i, img in enumerate(images):

            if images_fname is not None:
                fname = '%s' % images_fname[i]
            else:
                fname = '%d.png' % i
            hu.save_image(os.path.join(images_dir, fname), img)
        # if verbose:
        #     print('> Saved "images" in %s' % os.path.split(images_dir)[-1])

    print("=" * 100)

def get_checkpoint(savedir, return_model_state_dict=False):
    chk_dict = {}

    # score list
    score_list_fname = os.path.join(savedir, 'score_list.pkl')
    if os.path.exists(score_list_fname):
        score_list = hu.load_pkl(score_list_fname)
    else:
        score_list = []

    chk_dict['score_list'] = score_list
    if len(score_list) == 0:
        chk_dict['epoch'] = 0
    else:
        chk_dict['epoch'] = score_list[-1]['epoch'] + 1

    model_state_dict_fname = os.path.join(savedir, 'model.pth')
    if return_model_state_dict:
        if os.path.exists(model_state_dict_fname):
            chk_dict['model_state_dict'] = hu.torch_load(
                model_state_dict_fname)
        else:
            chk_dict['model_state_dict'] = {}

    return chk_dict


def create_jupyter(savedir, return_model_state_dict=False):
    chk_dict = {}

    # score list
    score_list_fname = os.path.join(savedir, 'score_list.pkl')
    score_list = hu.load_pkl(score_list_fname)

    chk_dict['score_list'] = score_list
    if len(score_list) == 0:
        chk_dict['epoch'] = 0
    else:
        chk_dict['epoch'] = score_list[-1]['epoch'] + 1

    if return_model_state_dict:
        model_state_dict_fname = os.path.join(savedir, 'model.pth')
        chk_dict['model_state_dict'] = hu.torch_load(model_state_dict_fname)

    return chk_dict



import sys, os, pprint

path = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
sys.path.insert(0, path)

# from haven.haven_jobs import slurm_manager as sm

  
# from haven import haven_chk as hc
# from haven import haven_results as hr
from haven import haven_utils as hu
from haven import haven_examples as he
import torch
import torchvision
import tqdm
import pandas as pd
import pprint
import itertools
import os
import pylab as plt
import time
import numpy as np
from torch.utils.data import RandomSampler, DataLoader
print()

try:
  import job_configs
except:
  pass
# from haven_utils import file_utils




# Job status
# ===========
def get_job(job_id):
    """Get job information."""
    command = "scontrol show job %s" % job_id
    job_info = ''
    while True:
      try:
        job_info = hu.subprocess_call(command)
        job_info = job_info.replace('\n', '')
        job_info = {v.split('=')[0]:v.split('=')[1] for v in job_info.split(' ') if '=' in v }
      except:
        print("scontrol time out and retry now")
        time.sleep(1)
        continue
      break
    return job_info
    

def get_jobs(user_name):
    # account_id = hu.subprocess_call('eai account get').split('\n')[-2].split(' ')[0]
    ''' get the first 3 jobs'''
    command = "squeue --user=%s" % user_name
    while True:
      try:
        job_list = hu.subprocess_call(command)
        job_list = job_list.split('\n')
        job_list = [v.lstrip().split(" ")[0] for v in job_list[1:]]
        result = []
        for job_id in job_list:
          result.append(get_job(job_id))
      except:
        print("scontrol time out and retry now")
        time.sleep(1)
        continue
      break
    return result

           
# Job kill
# ===========
def kill_job(job_id):
    """Kill a job job until it is dead."""
    kill_command  = "scancel %s" % job_id
    while True:
      try:
        hu.subprocess_call(kill_command)     # no return message after scancel
      except:
        print("scancel time out and retry now")
        time.sleep(1)
        continue
      break
    return
import argparse


def get_existing_slurm_job_commands(exp_list, savedir_base):
    existing_job_commands = []
    for exp_dict in exp_list:
        exp_id = hu.hash_dict(exp_dict)
        savedir = os.path.join(savedir_base, exp_id)
        file_name = os.path.join(savedir, "job_dict.json")
        if not os.path.exists(file_name):
            continue
        job_dict = hu.load_json(file_name)
        job_id = job_dict["job_id"]
        job_status = hu.subprocess_call("scontrol show job %s" % job_id).split("JobState=")[1].split(" ")[0]
        if job_status == "RUNNING" or job_status == "PENDING":
            existing_job_commands += [job_dict["command"]]
        
    return existing_job_commands

def get_job_spec():
    # read slurm setting
    lines = "#! /bin/bash \n"
    lines += "#SBATCH --account=%s \n" % job_configs.ACCOUNT_ID
    for key in list(job_configs.JOB_CONFIG.keys()):
        lines += "#SBATCH --%s=%s \n" % (key, job_configs.JOB_CONFIG[key])
    return lines

def submit_job(command, savedir):
    # read slurm setting
    lines = "#! /bin/bash \n"
    # if job_config is not None:
    #     lines += "#SBATCH --account=%s \n" % job_configs.ACCOUNT_ID
    #     for key in list(job_config.keys()):
    #         lines += "#SBATCH --%s=%s \n" % (key, job_config[key])
    lines += "#SBATCH --account=%s \n" % job_configs.ACCOUNT_ID
    for key in list(job_configs.JOB_CONFIG.keys()):
      lines += "#SBATCH --%s=%s \n" % (key, job_configs.JOB_CONFIG[key])
    path_log = os.path.join(savedir, "logs.txt")
    path_err = os.path.join(savedir, "err.txt")
    lines += "#SBATCH --output=%s \n" % path_log
    lines += "#SBATCH --error=%s \n" % path_err
    
    lines += command

    file_name = os.path.join(savedir, "bash.sh")
    hu.save_txt(file_name, lines)
    # launch the exp
    submit_command = "sbatch %s" % file_name
    while True:
      try:
        job_id = hu.subprocess_call(submit_command).split()[-1]
      except:
        print("slurm time out and retry now")
        time.sleep(1)
        continue
      break

    # save the command and job id in job_dict.json
    job_dict = {
        "command": command,
        "job_id": job_id
    }
    hu.save_json(os.path.join(savedir, "job_dict.json"), job_dict)

    # delete the bash.sh
    os.remove(file_name)

    return job_id

# 1. define the training and validation function
def trainval(exp_dict, savedir, args):
    """
    exp_dict: dictionary defining the hyperparameters of the experiment
    savedir: the directory where the experiment will be saved
    args: arguments passed through the command line
    """
    # 2. Create data loader and model 
    train_loader = he.get_loader(name=exp_dict['dataset'], split='train', 
                                 datadir=os.path.dirname(savedir),
                                 exp_dict=exp_dict)
    model = he.get_model(name=exp_dict['model'], exp_dict=exp_dict)

    # 3. load checkpoint
    chk_dict = hw.get_checkpoint(savedir)

    # 4. Add main loop
    for epoch in tqdm.tqdm(range(chk_dict['epoch'], 10), 
                           desc="Running Experiment"):
        # 5. train for one epoch
        train_dict = model.train_on_loader(train_loader, epoch=epoch)

        # 6. get and save metrics
        score_dict = {'epoch':epoch, 'acc': train_dict['train_acc'], 
                      'loss':train_dict['train_loss']}
        chk_dict['score_list'] += [score_dict]

        images = model.vis_on_loader(train_loader)

    hw.save_checkpoint(savedir, score_list=chk_dict['score_list'], images=[images])
    print('Experiment done\n')

if __name__ == "__main__":
  # task 1 - submit example job
  """
  Run echo 35 and forward the logs to <savedir>/logs.txt
  """
  # command = 'echo 35'
  # savedir = '/home/xhdeng/shared/results/test_slurm/example_get_jobs'
  # job_id = submit_job(command, savedir)

  # # task 2 - get job info as dict
  # """
  # Get job info as dict and save it as json in the directory specified below
  # """
  # hu.save_json('/home/xhdeng/shared/results/test_slurm/example/job_info.json', job_info)

  # task 3 - kill job
  # """
  # Kill job then gets its info as dict and save it as json in the directory specified below
  # it should say something like CANCELLED
  # """
  # kill_job(job_id)
  # job_info = get_job(job_id)
  # hu.save_json('/home/xhdeng/shared/results/test_slurm/example_kill_job/job_info_dead.json', job_info)
  
  # task 4 - get all jobs from an account as a list
  # """  
  # Get all jobs from an account as a list and save it in directory below
  # """
  # job_info_list = get_jobs(job_configs.USER_NAME)
  # hu.save_json('/home/xhdeng/shared/results/test_slurm/example_get_jobs/job_info_list.json', job_info_list)

  # # task 5 - run 10 jobs using threads
  # """
  # Use thee parallel threads from Haven and run these jobs in parallel
  # """
  # pr = hu.Parallel()

  # for i in range(1,20):
  #   command = 'echo %d' % i
  #   savedir = '/home/xhdeng/shared/results/test_slurm/example_%d' % i
    
  #   pr.add(submit_job, command, savedir)

  # pr.run()
  # pr.close()

  # # task 6 with menu - run mnist experiments on these 5 learning rates
  import argparse

  parser = argparse.ArgumentParser()

  parser.add_argument('-ei', '--exp_id')
  args = parser.parse_args()
  exp_list = []
  for lr in [1, 1e-1, 1e-2, 1e-3, 1e-4, 1e-5, 'bug']:
      exp_list += [{'lr':lr, 'dataset':'mnist', 'model':'linear'}]
  
  if args.exp_id is None:
    # run jobs
    
    print("\nTotal Experiments:", len(exp_list))
    prompt = ("\nMenu:\n"
              "  0)'ipdb' run ipdb for an interactive session; or\n"
              "  1)'reset' to reset the experiments; or\n"
              "  2)'run' to run the remaining experiments and retry the failed ones; or\n"
              "  3)'status' to view the job status; or\n"
              "  4)'kill' to kill the jobs.\n"
              "Type option: "
              )
    
    option = input(prompt)
    if option == 'run':
      # only run if job has failed or never ran before
      
      for exp_dict in exp_list:
        exp_id = hu.hash_dict(exp_dict)
        command = 'python test_slurm.py -ei %s' % exp_id

    elif option == 'reset':
      # ressset each experiment (delete the checkpoint and reset)
      command = 'python test_slurm.py -ei %s' % exp_id
      for exp_dict in exp_list:
        exp_id = hu.hash_dict(exp_dict)
        command = 'python test_slurm.py -ei %s' % exp_id

    elif option == 'status':
      # get job status of each exp
      for exp_dict in exp_list:
        exp_id = hu.hash_dict(exp_dict)
    elif option == 'kill':
      # make sure all jobs for the exps are dead
      for exp_dict in exp_list:
        exp_id = hu.hash_dict(exp_dict)

  else:
    for exp_dict in exp_list:
      exp_id = hu.hash_dict(exp_dict)
      savedir = '/home/xhdeng/shared/results/test_slurm/%s' % exp_id
      if exp_id is not None and exp_id == args.exp_id:
        trainval(exp_dict, savedir, args={})
   
  

 



import sys, os, pprint
path = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
sys.path.insert(0, path)

import sys, os, pprint
from haven.haven_results import plots_line as pl
from haven import haven_results as hr
from haven import haven_utils as hu
from haven.haven_jobs import slurm_manager as sm

import pprint
import pandas as pd

def get_latex_table(score_df, 
                    columns=None,
                    rows=None,
                    filter_dict=dict(), 
                    map_row_dict_dict=dict(), 
                    map_col_dict=dict(), **kwargs):

    # break it
    dicts = score_df.T.to_dict()

    dicts_new = {}
    for i in dicts:
        exp_score_dict = dicts[i]
        row_label = '_'.join([map_row_dict_dict.get(exp_score_dict[l], exp_score_dict[l]) for l in exp_score_dict if l in rows])
        col_scores = {k:v for k, v in exp_score_dict.items() if k in columns}
        dicts_new[row_label] = col_scores
    
    df_new = pd.DataFrame(dicts_new).T
    return df_new.to_latex(**kwargs)



if __name__ == "__main__":
    savedir_base = '.tmp'
    exp_dict = {'model':{'name':'mlp', 'n_layers':30}, 
                'dataset':'mnist', 'batch_size':1}


    score_list = [{'epoch': 4, 'acc':0.5}, {'epoch': 6, 'acc':0.9}]

    hu.save_pkl(os.path.join(savedir_base, hu.hash_dict(exp_dict),
                    'score_list.pkl'), score_list)
                    
    hu.save_json(os.path.join(savedir_base, hu.hash_dict(exp_dict),
                    'exp_dict.json'), exp_dict)

    exp_dict = {'model':{'name':'mlp2', 'n_layers':35}, 
                'dataset':'mnist', 'batch_size':1}
    score_list = [{'epoch': 2, 'acc':0.1}, {'epoch': 6, 'acc':0.3}]
    hu.save_pkl(os.path.join(savedir_base, hu.hash_dict(exp_dict),
                    'score_list.pkl'), score_list)
    hu.save_json(os.path.join(savedir_base, hu.hash_dict(exp_dict),
                    'exp_dict.json'), exp_dict)
    # check if score_list can be loaded and viewed in pandas
    exp_list = hu.get_exp_list(savedir_base=savedir_base)
    score_df = hr.get_score_df(exp_list, savedir_base=savedir_base)


    print(get_latex_table(score_df, columns=['acc', 'epoch'], rows=['model.name'], filter_dict=dict(), 
                          caption='Test',  float_format='%.2f', label='ref:table1',
            map_row_dict_dict=dict(), map_col_dict=dict()))
    print()


import torch, torchvision, os, pprint
import tqdm
import argparse, pandas as pd
import numpy as np
from torch.utils.data import TensorDataset
from haven import haven_utils as hu
from haven import haven_wizard as hw

def get_loader(name, split, datadir, exp_dict):
    if name == 'syn':
        # get dataset and loader
        X = torch.randn(5000, 1, 28, 28)
        y = torch.randint(0, 2, (5000,1))
        dataset = TensorDataset(X, y)
        loader = torch.utils.data.DataLoader(dataset, batch_size=256)

    if name == 'mnist':
        # get dataset and loader
        train = True if split == 'train' else False
        transform = torchvision.transforms.Compose([torchvision.transforms.ToTensor()])
        dataset = torchvision.datasets.MNIST(root=datadir, 
                                             train=train, download=True,
                                             transform=transform)
        loader = torch.utils.data.DataLoader(dataset, batch_size=256)

    if name == 'pascal_small':
        # get dataset and loader
        train = True if split == 'train' else False
        transform = torchvision.transforms.Compose([torchvision.transforms.ToTensor()])
        dataset = torchvision.datasets.MNIST(root=datadir, 
                                             train=train, download=True,
                                             transform=transform)
        loader = torch.utils.data.DataLoader(dataset, batch_size=256)


    return loader

def get_model(name, exp_dict):
    if name == 'linear':
        model = Linear()
        model.opt = torch.optim.Adam(model.parameters(), lr=exp_dict['lr'])

    if name == 'mlp':
        model = Mlp()
        model.opt = torch.optim.Adam(model.parameters(), lr=exp_dict['lr'])

    if name == 'seg_model':
        model = Mlp()
        model.opt = torch.optim.Adam(model.parameters(), lr=exp_dict['lr'])


    return model

# =====================================================
class Linear(torch.nn.Module):
    def __init__(self):
        super().__init__()
        self.model = torch.nn.Linear(784, 10)

    def train_on_batch(self, batch, **extras):
        self.opt.zero_grad()
        self.train()

        images, labels = batch
        logits = self.model.forward(images.view(images.shape[0], -1))
        criterion = torch.nn.CrossEntropyLoss(reduction="mean")
        loss = criterion(logits, labels.view(-1))
        loss.backward()

        self.opt.step()

        return {"train_loss": loss.item(), 'train_acc':(logits.argmax(dim=1) == labels).float().mean().item()}

    @torch.no_grad()
    def vis_on_loader(self, loader, **extras):
        self.eval()

        for batch in loader:
            images, labels = batch
            probs = torch.softmax(self.model.forward(images.view(images.shape[0], -1)), dim=1)
            score, label = probs.max(dim=1)
            i_list = []
            for i in range(probs.shape[0]):
                pil_img = hu.save_image('tmp', images[i], return_image=True)
                img = get_image(pil_img, 'Predicted %d (Prob: %.2f)' % (label[i], score[i]))
                i_list += [img]
                if i > 5: 
                    break
            return np.hstack(i_list)[:, :, None].repeat(3, axis=2)
    
    def train_on_loader(self, loader, **extras):
        for batch in tqdm.tqdm(loader, desc="Epoch %d" % extras.get('epoch'), 
                               leave=False):
            train_dict = self.train_on_batch(batch)

        return train_dict

# MLP
class Mlp(torch.nn.Module):
    def __init__(self, input_size=784,
                 hidden_sizes=[512, 256],
                 n_classes=10,
                 bias=True, dropout=False):
        super().__init__()
        self.input_size = input_size
        self.hidden_layers = torch.nn.ModuleList([torch.nn.Linear(in_size, out_size, bias=bias) for
                                            in_size, out_size in zip([self.input_size] + hidden_sizes[:-1], hidden_sizes)])
        self.output_layer = torch.nn.Linear(hidden_sizes[-1], n_classes, bias=bias)

    def train_on_batch(self, batch, **extras):
        self.opt.zero_grad()
        self.train()

        images, labels = batch
        logits = self.forward(images)
        criterion = torch.nn.CrossEntropyLoss(reduction="mean")
        loss = criterion(logits, labels.view(-1))
        loss.backward()

        self.opt.step()

        return {"train_loss": loss.item(), 'train_acc':(logits.argmax(dim=1) == labels).float().mean().item()}

    def train_on_loader(self, loader, **extras):
        for batch in tqdm.tqdm(loader, desc="Epoch %d" % extras.get('epoch'), 
                               leave=False):
            train_dict = self.train_on_batch(batch)

        return train_dict

    def forward(self, x):
        x = x.view(-1, self.input_size)
        out = x
        for layer in self.hidden_layers:
            Z = layer(out)
            out = torch.nn.functional.relu(Z)

        logits = self.output_layer(out)

        return logits

    def get_state_dict(self):
        state_dict = {"model": self.state_dict(),
                      "opt":self.opt.state_dict()}

        return state_dict
        
    def set_state_dict(self, state_dict):
        self.load_state_dict(state_dict["model"])
        self.opt.load_state_dict(state_dict["opt"])


def get_image(pil_img, title):
    import pylab as plt
    from PIL import Image
    # plt.figure(figsize=(16,10))
    plt.imshow(pil_img)
    ax = plt.gca()
    plt.title(title, fontsize=20)
    plt.axis('off')
    plt.tight_layout()
    plt.savefig('tmp.png', bbox_inches='tight', pad_inches=0)
    plt.close()
    return np.array(Image.open('tmp.png').convert('L'))


def save_example_results(savedir_base='results'):
    import os, pandas
    import requests, io
    import matplotlib.pyplot as plt 

    from .. import haven_results as hr
    from .. import haven_utils as hu
    from PIL import Image

    # create hyperparameters
    exp_list = [{'dataset':'mnist', 'model': 'mlp',
                'lr':lr} for lr in [1e-1, 1e-2, 1e-3]]

    for i, exp_dict in enumerate(exp_list):
        # get hash for experiment
        exp_id = hu.hash_dict(exp_dict)
        
        # add scores for loss, and accuracy
        score_list = []
        for e in range(1,10):
            score_list += [{'epoch':e, 
                            'loss':1 - e*exp_dict['lr'] * 0.9, 
                            'acc':e*exp_dict['lr'] * 0.1}]
        # save scores and images
        hu.save_json(os.path.join(savedir_base, exp_id, 'exp_dict.json'), exp_dict)
        hu.save_pkl(os.path.join(savedir_base, exp_id, 'score_list.pkl'), score_list)
        
        url = 'https://raw.githubusercontent.com/haven-ai/haven-ai/master/haven/haven_examples/data/%d.png' % (i+1)
        response = requests.get(url).content
        img = plt.imread(io.BytesIO(response), format='JPG')
        hu.save_image(os.path.join(savedir_base, exp_id, 'images/1.png'), img[:,:,:3])


import sys, os, pprint

path = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
sys.path.insert(0, path)

from haven import haven_examples as he

if __name__ == "__main__":
    he.save_example_results(savedir_base='.tmp')



from .. import haven_utils
from .. import haven_results as hr
from .. import haven_utils as hu 
from .. import haven_share as hd

import os
import pprint, json
import copy
import pprint
import pandas as pd 
from . import widgets as wdg
try:
    import ast
    from ipywidgets import Button, HBox, VBox
    from ipywidgets import widgets

    from IPython.display import display
    from IPython.core.display import Javascript, display, HTML
    from IPython.display import FileLink, FileLinks
    from ipywidgets.widgets.interaction import show_inline_matplotlib_plots
except:
    print('widgets not available...')


def tables_tab(db, output):
    w_columns = wdg.SelectMultiple(header="Hyperparameters:", 
                            options=db.rm.exp_params,
                            db_vars=db.vars, 
                            var='columns', select_all=True)
    w_score_columns = wdg.SelectMultiple(header="Metrics:", 
                            options=db.rm.score_keys,
                            db_vars=db.vars, 
                            var='score_columns', select_all=True)


    bstatus = widgets.Button(description="Jobs Status")
    blogs = widgets.Button(description="Jobs Logs")
    bfailed = widgets.Button(description="Jobs Failed")

    b_table = widgets.Button(description="Display Table")
    b_meta = widgets.Button(description="Display Meta Table")
    b_diff = widgets.Button(description="Display Filtered Table")

    w_avg_across = wdg.Dropdown(header='Avg Across',
                                options=['None'] + db.rm.exp_params,
                            db_vars=db.vars, 
                            var='avg_across')

    button = widgets.VBox([ 
                            widgets.HBox([w_columns.get_widget(), w_score_columns.get_widget(), w_avg_across.get_widget()]),
                            widgets.HBox([b_table, bstatus, blogs, bfailed, ]),
                           
    ])
    output_plot = widgets.Output()

    with output:
        display(button)
        display(output_plot)

    def on_table_clicked(b):
        output_plot.clear_output()
        with output_plot:
            db.update_rm()

            score_table = db.rm.get_score_table(columns=w_columns.update(), 
                                            score_columns=w_score_columns.update(),
                                            avg_across=w_avg_across.update())
            display(score_table) 

    def on_job_status_clicked(b):
        output_plot.clear_output()
        with output_plot:
            db.update_rm()
            summary_list = db.rm.get_job_summary(verbose=db.rm.verbose,
                                               add_prefix=True)
            summary_dict = hu.group_list(summary_list, key='job_state', return_count=True)
            display(summary_dict)

            summary_dict = hu.group_list(summary_list, key='job_state', return_count=False)

            for state in summary_dict:
                n_jobs = len(summary_dict[state])
                if n_jobs:
                    display('Experiments %s: %d' %(state, n_jobs))
                    df = pd.DataFrame(summary_dict[state])
                    display(df.head())

    def on_logs_clicked(b):
        output_plot.clear_output()
        with output_plot:
            summary_list = db.rm.get_job_summary(verbose=db.rm.verbose,
                                               add_prefix=True)
            
            n_logs = len(summary_list)
        
            for i, logs in enumerate(summary_list):
                print('\nLogs %d/%d' % (i+1, n_logs), '='*50)
                print('exp_id:', logs['exp_id'])
                print('job_id:', logs['job_id'])
                print('job_state:', logs['job_state'])
                print('savedir:', os.path.join(db.rm_original.savedir_base, logs['exp_id']))

                print('\nexp_dict')
                print('-'*50)
                pprint.pprint(logs['exp_dict'])
                
                print('\nLogs')
                print('-'*50)
                pprint.pprint(logs.get('logs'))     
    
    def on_failed_clicked(b):
        output_plot.clear_output()
        with output_plot:
            db.update_rm()
            summary_list = db.rm.get_job_summary(verbose=db.rm.verbose,
                                               add_prefix=True)
            summary_dict = hu.group_list(summary_list, key='job_state', return_count=False)
            if 'FAILED' not in summary_dict:
                display('NO FAILED JOBS')
                return
            n_failed = len(summary_dict['FAILED'])
        
            if n_failed == 0:
                display('no failed experiments')
            else:
                for i, failed in enumerate(summary_dict['FAILED']):
                    print('\nFailed %d/%d' % (i+1, n_failed), '='*50)
                    print('exp_id:', failed['exp_id'])
                    print('job_id:', failed['job_id'])
                    print('job_state:', 'FAILED')
                    print('savedir:', os.path.join(db.rm_original.savedir_base, failed['exp_id']))

                    print('\nexp_dict')
                    print('-'*50)
                    pprint.pprint(failed['exp_dict'])
                    
                    print('\nLogs')
                    print('-'*50)
                    pprint.pprint(failed.get('logs'))

    # Add call listeners
    b_table.on_click(on_table_clicked)
    bstatus.on_click(on_job_status_clicked)
    blogs.on_click(on_logs_clicked)
    bfailed.on_click(on_failed_clicked)






from pkg_resources import get_distribution                                        
                                                                                  
__version__ = get_distribution('haven-ai').version


from . import haven_utils as hu
from . import haven_chk as hc
import os
import time
import copy
import pandas as pd
import numpy as np 
import getpass
import pprint
import requests
try:
    import borgy_job_service_client
except:
    pass

# Api
# ==============
def get_api(**kwargs):
    # Get Borgy API
    config = borgy_job_service_client.Configuration()
    config.host = "https://job-service.borgy.elementai.net"

    api_client = borgy_job_service_client.ApiClient(config)

    api_client = borgy_job_service_client.ApiClient(config)
    api_client.set_default_header('X-User', kwargs['username'])

    # create an instance of the API class
    return borgy_job_service_client.JobsApi(api_client)


# Job submission
# ==============
def submit_job(api, account_id, command, job_config, workdir, savedir_logs=None):
    eai_command = get_borgy_command(job_config, command, savedir_logs, workdir)
    job_id = hu.subprocess_call(eai_command).replace("\n", "")

    return job_id


def get_borgy_command(job_config, command, savedir, workdir):
    """Compose the borgy submit command."""
    eai_command = "borgy submit "

    # Add the Borgy options
    for k, v in job_config.items():
        if k in ['username']:
            continue
        # Handle all the different type of parameters
        if k == "restartable":
            eai_command += " --%s" % k
        elif k in ["gpu", "cpu", "mem"]:
            eai_command += " --%s=%s" % (k, v)
        elif isinstance(v, list):
            for v_tmp in v:
                eai_command += " --%s %s" % (k, v_tmp)
        elif k in ["userid", "email"]:
            continue
        else:
            eai_command += " --%s %s" % (k, v)
    eai_command += " --workdir %s" % workdir

    # Add the python command to run and the logs file
    if savedir:
        path_log = os.path.join(savedir, "logs.txt")
        path_err = os.path.join(savedir, "err.txt")
        command = '"%s 1>%s 2>%s"' % (command, path_log, path_err)

    eai_command += " -- /bin/bash -c %s" % command

    # Return the Borgy command in Byte format
    return r'''body'''.replace("body", eai_command)


# Job status
# ===========
def get_jobs_dict(api, job_id_list, query_size=20):
    # get jobs
    jobs = []
    for i in range(0, len(job_id_list), query_size):
        job_id_string = "id IN ("
        for job_id in  job_id_list[i:i + query_size]:
            job_id_string += "'%s', " % job_id
        job_id_string = job_id_string[:-2] + ")"
        jobs += api.v1_jobs_get(q=job_id_string)

    jobs_dict = {job.id: job for job in jobs}

    return jobs_dict

def get_job(api, job_id):
    """Get a Borgy job."""
    return api.v1_jobs_job_id_get(job_id)

def get_jobs(api, username):
    return api.v1_jobs_get(
            q="alive=true AND name='{}' "
            "ORDER BY createdOn "
            "DESC LIMIT 1000".format(username))
           

# Job kill
# ===========
def kill_job(api, job_id):
    """Kill a job job until it is dead."""
    job = get_job(api, job_id)

    if not job.alive:
        print('%s is already dead' % job_id)
    else:
        hu.subprocess_call("borgy kill %s" % job_id)
        print('%s CANCELLING...' % job_id)
        job = get_job(api, job_id)
        while job.state == "CANCELLING":
            time.sleep(2.0)
            job = get_job(api, job_id)

        print('%s now is dead.' % job_id)


import sys, os, pprint

path = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
sys.path.insert(0, path)

import wandb
from haven import haven_wizard as hw

if __name__ == "__main__":
  # first way
  score_dict = {'loss':loss}
  wandb.send(score_dict)
    
  # second way
  chk = load_checkpoint(savedir)
  hw.save_checkpoint(savedir, score_dict=score_dict, wandb_config={})



import contextlib
import copy
import hashlib
import itertools
import json
import tqdm

import pprint
import os
import pickle
import shlex
import subprocess
import threading
import time
import numpy as np
import pylab as plt
import torch

from .. import haven_img as hi
from .image_utils import *
from .file_utils import *
from .string_utils import *
from .exp_utils import *
from datetime import datetime
from PIL import Image


def get_function_from_file():
    pass

def create_command(base_command, args):
    """
    args is the parser
    """
    run_command = base_command
    arg_keys= vars(args).keys()

    assert('exp_group_list' in arg_keys)
    assert('exp_id' in arg_keys)
    assert('run_jobs' in arg_keys)

    for a, v in vars(args).items():
        if a == 'exp_group_list' or a == 'exp_id' or a == 'run_jobs' or a == 'reset':
            print('argument: %s ignored...' % a)
            continue

        run_command += ' --%s %s' % (a, v)
    print('command: %s' % run_command)
    return run_command 

def mask_on_image(mask, image):
    from skimage.color import label2rgb
    from skimage.color import color_dict, colorlabel
    from skimage.segmentation import mark_boundaries
    default_colors = ['red', 'blue', 'yellow', 'magenta', 'green',
                     'indigo', 'darkorange', 'cyan', 'pink', 'yellowgreen']

    mask = mask.squeeze().astype(int)
    image = hi.image_as_uint8(image) / 255.
    labels = [l for l in np.unique(mask) if l < len(color_dict)]
    colors =  (default_colors + list(color_dict.keys())[len(default_colors):])
    colors =  np.array(colors)[labels]
    
    image_label_overlay = label2rgb(mask, image=f2l(image).squeeze().clip(0,1), 
                                    colors=colors, bg_label=0, bg_color=None, kind='overlay')
    return mark_boundaries(image_label_overlay, mask)

def get_image(img, denorm=None, size=None, points=None, radius=10,
              mask=None, heatmap=None):
    return save_image(None, img, denorm, size, points, radius,
               mask, heatmap, return_image=True)

def save_image(fname, img, denorm=None, size=None, points=None, radius=10,
               mask=None, heatmap=None, makedirs=True, return_image=False):
    """Save an image into a file.

    Parameters
    ----------
    fname : str
        Name of the file
    img : [type]
        Image data. #TODO We asume it is.....?????? \in [0, 1]? Numpy? PIL? RGB?
    makedirs : bool, optional
        If enabled creates the folder for saving the file, by default True
    """
    if denorm:
        img = denormalize(img, mode=denorm)
    if points is not None:
        if isinstance(img, np.ndarray):
            img = torch.FloatTensor(img)
        img = img.squeeze()
        if img.ndim == 2:
            img = img[None].repeat(3,1,1)
        y_list, x_list = np.where(points.squeeze())
        c_list = []
        for y, x in zip(y_list, x_list):
            c_list += [points.squeeze()[y, x]]
        img = hi.points_on_image(y_list, x_list, img, 
                 radius=radius, c_list=c_list)

    if mask is not None:
        img = mask_on_image(mask, img)

    

    if img.dtype == 'uint8':
        img = Image.fromarray(img)
    else:
        arr = f2l(t2n(img)).squeeze()
        # print(arr.shape)
        if size is not None:  
            arr = Image.fromarray(arr)
            arr = arr.resize(size)
            arr = np.array(arr)

        img = Image.fromarray(np.uint8(arr * 255))
    if return_image:
        return img

    if fname is not None:
        dirname = os.path.dirname(fname)
        if makedirs and dirname != '':
            os.makedirs(dirname, exist_ok=True)
        img.save(fname)


def load_txt(fname):
    """Load the content of a txt file.

    Parameters
    ----------
    fname : str
        File name

    Returns
    -------
    list
        Content of the file. List containing the lines of the file
    """
    with open(fname, "r", encoding="utf-8") as f:
        lines = f.readlines()
    return lines

def save_txt(fname, lines):
    """Load the content of a txt file.

    Parameters
    ----------
    fname : str
        File name

    Returns
    -------
    list
        Content of the file. List containing the lines of the file
    """
    with open(fname, "w", encoding="utf-8") as f:
        for l in lines:
            f.writelines(l)
    

def torch_load(fname, map_location=None):
    """Load the content of a torch file.

    Parameters
    ----------
    fname : str
        File name
    map_location : [type], optional
        Maping the loaded model to a specific device (i.e., CPU or GPU), this
        is needed if trained in CPU and loaded in GPU and viceversa, by default
        None

    Returns
    -------
    [type]
        Loaded torch model
    """
    obj = torch.load(fname, map_location=map_location)

    return obj


def torch_save(fname, obj):
    """Save data in torch format.

    Parameters
    ----------
    fname : str
        File name
    obj : [type]
        Data to save
    """
    # Create folder
    os.makedirs(os.path.dirname(fname), exist_ok=True)  # TODO: add makedirs parameter?

    # Define names of temporal files
    fname_tmp = fname + ".tmp"  # TODO: Make the safe flag?

    torch.save(obj, fname_tmp)
    if os.path.exists(fname):
        os.remove(fname)
    os.rename(fname_tmp, fname)


class Parallel:
    """Class for run a function in parallel."""
    def __init__(self):
        """Constructor"""
        self.threadList = []
        self.count = 0
        self.thread_logs = []

    def add(self, func,  *args):
        """Add a function to run as a process.

        Parameters
        ----------
        func : function
            Pointer to the function to parallelize
        args : list
            Arguments of the funtion to parallelize
        """
        self.threadList += [
            threading.Thread(target=func, name="thread-%d" % self.count,
                             args=args)]
        self.count += 1

    def run(self):
        """Run the added functions in parallel"""
        for thread in tqdm.tqdm(self.threadList, desc='Starting threads', leave=False):
            thread.daemon = True
            thread.start()

    def close(self):
        """Finish: wait for all the functions to finish"""
        for thread in tqdm.tqdm(self.threadList, desc='Joining threads', leave=False):
            thread.join()


def pretty_print_df(df):
    # wrap text for prettiness
    for c in df.columns:
        if df[c].dtype == 'O':
            # df[c] = df[c].str.wrap(wrap_size)
            df[c] = df[c].apply(pprint.pformat)
    return df 

def flatten_column(result_dict):
    new_dict = {}

    for k, v in result_dict.items():
        new_dict.update(flatten_dict(k, v))

    result_dict = new_dict
    return result_dict

def sort_df_columns(table, also_first=[]):
    first = ['exp_id', 'job_state', 'job_id',  'restarts', 'started_at']
    first += also_first
    col_list = []
    for col in first:
        if col in table.columns:
            col_list += [col]
    for col in table.columns:
        if col in first:
            continue
        col_list += [col]
        
    return table[col_list]

def subprocess_call(cmd_string):
    """Run a terminal process.

    Parameters
    ----------
    cmd_string : str
        Command to execute in the terminal

    Returns
    -------
    [type]
        Error code or 0 if no error happened
    """
    return subprocess.check_output(
        shlex.split(cmd_string), shell=False).decode("utf-8")


def copy_code(src_path, dst_path, verbose=1):
    """Copy the code.
    
    Typically, when you run an experiment, first you copy the code used to the
    experiment folder. This function copies the code using rsync terminal
    command.

    Parameters
    ----------
    src_path : str
        Source code directory
    dst_path : str
        Destination code directory
    verbose : int, optional
        Verbosity level. If 0 does not print stuff, by default 1

    Raises
    ------
    ValueError
        [description]
    """
    time.sleep(.5)  # TODO: Why? Why?

    if verbose:
        print("  > Copying code from %s to %s" % (src_path, dst_path))

    # Create destination folder
    os.makedirs(dst_path, exist_ok=True)

    # Define the command for copying the code using rsync
    rsync_code = "rsync -av -r -q  --delete-before --exclude='.tmp/' --exclude='.git/' " \
                 " --exclude='*.pyc' --exclude='__pycache__/' %s %s" % (
                     src_path, dst_path)

    # Run the command in the terminal
    try:
        subprocess_call(rsync_code)
    except subprocess.CalledProcessError as e:
        raise ValueError("Ping stdout output:\n", e.output)

    time.sleep(.5)  # TODO: Why?


def zipdir(src_dirname, out_fname, include_list=None):
    """Compress a folder using ZIP.

    Parameters
    ----------
    src_dirname : str
        Directory to compress
    out_fname : str
        File name of the compressed file
    include_list : list, optional
        List of files to include. If None, include all files in the folder, by
        default None
    """
    import zipfile  # TODO: Move to the beggining of the file
    # TODO: Do we need makedirs?
    # Create the zip file
    zipf = zipfile.ZipFile(out_fname, 'w', zipfile.ZIP_DEFLATED)

    # ziph is zipfile handle
    for root, dirs, files in os.walk(src_dirname):
        for file in files:
            # Descard files if needed
            if include_list is not None and file not in include_list:
                continue
            
            abs_path = os.path.join(root, file)
            rel_path = fname_parent(abs_path)  # TODO: fname_parent not defined
            print(rel_path)
            zipf.write(abs_path, rel_path)

    zipf.close()


def zip_score_list(exp_list, savedir_base, out_fname, include_list=None):
    """Compress a list of experiments in zip.

    Parameters
    ----------
    exp_list : list
        List of experiments to zip
    savedir_base : str
        Directory where the experiments from the list are saved
    out_fname : str
        File name for the zip file
    include_list : list, optional
        List of files to include. If None, include all files in the folder, by
        default None
    """
    for exp_dict in exp_list:  # TODO: This will zip only the last experiments, zipdir will overwritwe the previous file
        # Get the experiment id
        exp_id = hash_dict(exp_dict)
        # Zip folder
        zipdir(os.path.join(savedir_base, exp_id),
               out_fname, include_list=include_list)


def time_to_montreal(fname=None, timezone='US/Eastern'):
    """Get time in Montreal zone.

    Returns
    -------
    str
        Current date at the selected timezone in string format
    """
    # Get time
    os.environ['TZ'] = timezone
    time.tzset()
    if fname:
        tstamp = os.path.getctime(fname)
    else:
        tstamp = time.time()
        
    time_str = datetime.fromtimestamp(tstamp).strftime('%I:%M %p (%b %d)')
    
    return time_str


def time2mins(time_taken):
    """Convert time into minutes.

    Parameters
    ----------
    time_taken : float
        Time in seconds

    Returns
    -------
    float
        Minutes
    """
    return time_taken / 60.


def n2t(x, dtype="float"):  # TODO: dtype is not used!!
    """Array or Numpy array to Pytorch tensor.

    Parameters
    ----------
    x : array or Numpy array
        Data to transform
    dtype : [type]
        [description]

    Returns
    -------
    Pytorch tensor
        x converted to pytorch tensor format
    """
    if isinstance(x, (int, np.int64, float)):
        x = np.array([x])

    if isinstance(x, np.ndarray):
        x = torch.from_numpy(x)
    return x


def t2n(x):
    """Pytorch tensor to Numpy array.

    Parameters
    ----------
    x : Pytorch tensor
        A Pytorch tensor to transform

    Returns
    -------
    Numpy array
        x transformed to numpy array
    """
    try:
        x = x.detach().cpu().numpy()
    except:
        x = x

    return x


def l2f(X):
    """Move the channels from the last dimension to the first dimension.

    Parameters
    ----------
    X : Numpy array
        Tensor with the channel dimension at the last dimension

    Returns
    -------
    Numpy array
        X transformed with the channel dimension at the first dimension
    """
    if X.ndim == 3 and (X.shape[0] == 3 or X.shape[0] == 1):
        return X
    if X.ndim == 4 and (X.shape[1] == 3 or X.shape[1] == 1):
        return X

    if X.ndim == 4 and (X.shape[1] < X.shape[3]):
        return X

    # Move the channel dimension from the last position to the first one
    if X.ndim == 3:
        return np.transpose(X, (2, 0, 1))
    if X.ndim == 4:
        return np.transpose(X, (0, 3, 1, 2))

    return X


def f2l(X):
    """Move the channels from the first dimension to the last dimension.

`   Parameters
    ----------
    X : Numpy array
        Tensor with the channel dimension at the first dimension

    Returns
    -------
    Numpy array
        X transformed with the channel dimension at the last dimension
    """
    if X.ndim == 3 and (X.shape[2] == 3 or X.shape[2] == 1):
        return X
    if X.ndim == 4 and (X.shape[3] == 3 or X.shape[3] == 1):
        return X

    # Move the channel dimension from the first position to the last one
    if X.ndim == 3:
        return np.transpose(X, (1, 2, 0))
    if X.ndim == 4:
        return np.transpose(X, (0, 2, 3, 1))

    return X


def n2p(image):  #TODO: Create p2n function and use it in get_image()
    """Numpy image to PIL image.

    Parameters
    ----------
    image : Numpy array
        Input image in numpy format

    Returns
    -------
    PIL image
        Input image converted into PIL format
    """
    image = f2l(image.squeeze())
    if image.max() <= 1:
        image = image * 255
    return Image.fromarray(image.astype('uint8'))


def _denorm(image, mu, var, bgr2rgb=False):
    """Denormalize an image.

    Parameters
    ----------
    image : [type]
        Image to denormalize
    mu : [type]
        Mean used to normalize the image
    var : [type]
        Variance used to normalize the image
    bgr2rgb : bool, optional
        Whether to also convert from bgr 2 rgb, by default False

    Returns
    -------
    [type]
        Denormalized image
    """
    if image.ndim == 3:
        result = image * var[:, None, None] + mu[:, None, None]  # TODO: Is it variance or std?
        if bgr2rgb:
            result = result[::-1]
    else:
        result = image * var[None, :, None, None] + mu[None, :, None, None]
        if bgr2rgb:
            result = result[:, ::-1]
    return result


def denormalize(img, mode=0):  # TODO: Remove the default value or set to a valid number, complete documentation
    """Denormalize an image.

    Parameters
    ----------
    img : [type]
        Input image to denormalize
    mode : int or str, optional
        Predefined denormalizations, by default 0
        If 1 or 'rgb'... 
        If 2 or 'brg'...,
        If 3 or 'basic'...
        Else do nothing

    Returns
    -------
    [type]
        Denormalized image
    """
    # _img = t2n(img)
    # _img = _img.copy()
    image = t2n(img).copy().astype("float")

    if mode in [1, "rgb"]:
        mu = np.array([0.485, 0.456, 0.406])
        var = np.array([0.229, 0.224, 0.225])
        image = _denorm(image, mu, var)

    elif mode in [2, "bgr"]:
        mu = np.array([102.9801, 115.9465, 122.7717])
        var = np.array([1, 1, 1])
        image = _denorm(image, mu, var, bgr2rgb=True).clip(0, 255).round()

    elif mode in [3, "basic"]:
        mu = np.array([0.5, 0.5, 0.5])
        var = np.array([0.5, 0.5, 0.5])
        image = _denorm(image, mu, var)

    # TODO: Add a case for 0 or None and else raise an error exception.

    return image


# def get_image(imgs, mask=None, label=False, enlarge=0, gray=False, denorm=0,
#               bbox_yxyx=None, annList=None, pretty=False, pointList=None,
#               **options):  # TODO: Issam, can you document this?
#     """[summary]

#     Parameters
#     ----------
#     imgs : [type]
#         [description]
#     mask : [type], optional
#         [description], by default None
#     label : bool, optional
#         [description], by default False
#     enlarge : int, optional
#         [description], by default 0
#     gray : bool, optional
#         [description], by default False
#     denorm : int, optional
#         [description], by default 0
#     bbox_yxyx : [type], optional
#         [description], by default None
#     annList : [type], optional
#         [description], by default None
#     pretty : bool, optional
#         [description], by default False
#     pointList : [type], optional
#         [description], by default None

#     Returns
#     -------
#     [type]
#         [description]
#     """
#     # TODO: Comment these transformations and make sure they are correct. Difficult to follow.
#     imgs = denormalize(imgs, mode=denorm)
#     if isinstance(imgs, Image.Image):
#         imgs = np.array(imgs)
#     if isinstance(mask, Image.Image):
#         mask = np.array(mask)

#     imgs = t2n(imgs).copy()
#     imgs = l2f(imgs)

#     if pointList is not None and len(pointList):
#         h, w = pointList[0]["h"], pointList[0]["w"]
#         mask_points = np.zeros((h, w))
#         for p in pointList:
#             y, x = p["y"], p["x"]
#             mask_points[int(h*y), int(w*x)] = 1
#         imgs = maskOnImage(imgs, mask_points, enlarge=1)

#     if pretty or annList is not None:
#         imgs = pretty_vis(imgs, annList, **options)
#         imgs = l2f(imgs)

#     if mask is not None and mask.sum() != 0:
#         imgs = maskOnImage(imgs, mask, enlarge)

#     if bbox_yxyx is not None:
#         _, _, h, w = imgs.shape
#         mask = bbox_yxyx_2_mask(bbox_yxyx, h, w)
#         imgs = maskOnImage(imgs, mask, enlarge=1)

#     # LABEL
#     elif (not gray) and (label or imgs.ndim == 2 or
#                          (imgs.ndim == 3 and imgs.shape[0] != 3) or
#                          (imgs.ndim == 4 and imgs.shape[1] != 3)):

#         imgs = label2Image(imgs)

#         if enlarge:
#             imgs = zoom(imgs, 11)

#     # Make sure it is 4-dimensional
#     if imgs.ndim == 3:
#         imgs = imgs[np.newaxis]

#     return imgs


def show_image(fname):  # TODO: Why the input is a filename instead of an image?
    """Load and image from hard disk and plot it.

    Parameters
    ----------
    fname : str
        Name of an image to load and show
    """
    ncols = 1  # TODO: Magic numbers
    nrows = 1
    height = 12
    width = 12
    fig, axs = plt.subplots(nrows=nrows, ncols=ncols,
                            figsize=(ncols*width, nrows*height))
    if not hasattr(axs, 'size'):  #TODO: What is this?
        axs = [[axs]]

    for i in range(ncols):
        img = plt.imread(fname)
        axs[0][i].imshow(img)
        axs[0][i].set_axis_off()
        axs[0][i].set_title('%s' % (fname))

    plt.axis('off')
    plt.tight_layout()

    plt.show()


def shrink2roi(img, roi):
    """[summary]

    Parameters
    ----------
    img : [type]
        [description]
    roi : [type]
        [description]

    Returns
    -------
    [type]
        [description]
    """
    ind = np.where(roi != 0)

    y_min = min(ind[0])
    y_max = max(ind[0])

    x_min = min(ind[1])
    x_max = max(ind[1])

    return img[y_min:y_max, x_min:x_max]


@contextlib.contextmanager
def random_seed(seed):
    """[summary]

    Parameters
    ----------
    seed : [type]
        [description]
    """
    state = np.random.get_state()
    np.random.seed(seed)
    try:
        yield
    finally:
        np.random.set_state(state)


def is_subset(d1, d2, strict=False):
    """[summary]

    Parameters
    ----------
    d1 : [type]
        [description]
    d2 : [type]
        [description]

    Returns
    -------
    [type]
        [description]
    """
    flag = True
    for k in d1:
        v1, v2 = d1.get(k), d2.get(k)

        # if both are values
        if not isinstance(v2, dict) and not isinstance(v1, dict):
            if v1 != v2:
                flag = False

        # if both are dicts
        elif isinstance(v2, dict) and isinstance(v1, dict):
            flag = is_subset(v1, v2)

        # if d1 is dict and not d2
        elif isinstance(v1, dict) and not isinstance(v2, dict):
            flag = False

        # if d1 is not and d2 is dict
        elif not isinstance(v1, dict) and isinstance(v2, dict):
            flag = False

        if flag is False:
            break

    return flag


def as_double_list(v):
    """[summary]

    Parameters
    ----------
    v : [type]
        [description]

    Returns
    -------
    [type]
        [description]
    """
    if not isinstance(v, list) and not isinstance(v, np.ndarray):
        v = [v]

    if not isinstance(v[0], list) and not isinstance(v[0], np.ndarray):
        v = [v]

    return v

def ignore_duplicates(list_of_dicts):
    # ensure no duplicates in exp_list
    dict_list = []
    hash_list = set()
    for data_dict in list_of_dicts:
        dict_id = hash_dict(data_dict)
        if dict_id in hash_list:
            continue
        else:
            hash_list.add(dict_id)
            dict_list += [data_dict]
    return dict_list

def check_duplicates(list_of_dicts):
    # ensure no duplicates in exp_list
    hash_list = set()
    for data_dict in list_of_dicts:
        dict_id = hash_dict(data_dict)
        if dict_id in hash_list:
            raise ValueError('duplicated dictionary detected:\n%s' %pprint.pformat(data_dict))
        else:
            hash_list.add(dict_id)

def load_py(fname):
    """[summary]

    Parameters
    ----------
    fname : [type]
        [description]

    Returns
    -------
    [type]
        [description]
    """
    import sys
    from importlib import reload
    from importlib import import_module

    if not os.path.exists(fname):
        raise ValueError('%s not found...' % fname)

    sys.path.append(os.path.dirname(fname))

    name = os.path.split(fname)[-1].replace('.py','')
    module = import_module(name) 
    reload(module)
    sys.path.pop()

    return module


def get_exp_list_from_ids(exp_id_list, savedir_base):
    exp_list = [] 
    for exp_id in exp_id_list:
        exp_list += [load_json(os.path.join(savedir_base, exp_id, 
                        'exp_dict.json'))]
    return exp_list


def flatten_dict(key_name, v_dict):
    if not isinstance(v_dict, dict):
        return {key_name: v_dict}
    
    leaf_dict = {}
    for k in v_dict:
        if key_name != '':
            k_new = key_name + "." + k
        else:
            k_new = k
        leaf_dict.update(flatten_dict(key_name=k_new, v_dict=v_dict[k]))
        
    return leaf_dict


def get_diff_columns(df, min_threshold=2, max_threshold='auto'):
    df.reset_index()
    if max_threshold == 'auto':
        max_threshold = df.shape[0]
    
    if max_threshold < 0:
        max_threshold = df.shape[0] + max_threshold
    
    column_count = []
    
    for column in df.columns:
        _set = set([str(v) for v in df[column].values])
        column_count.append(len(_set))
    indices = np.arange(len(df.columns))
    
    column_count = np.array(column_count)
    
    indices = indices[(column_count >= min_threshold) & (column_count <= max_threshold)]
    diff_columns = [df.columns[i] for i in indices]
    
    return diff_columns


def collate_fn(batch, mode='list'):
    if mode == 'list':
        batch_dict = {}
        for k in batch[0]:
            batch_dict[k] = []
            for i in range(len(batch)):
                
                batch_dict[k] += [batch[i][k]]

        return batch_dict 
        
    elif mode == 'default':
        return torch.utils.data.dataloader.default_collate(batch)

def timeit(func, n_times=10, **args):
    for i in range(n_times):
        if i ==1:
            s = time.time()
        func(**args)

    print('time:', (time.time()-s)/(n_times-1))



from .. import haven_utils
from .. import haven_results as hr
from .. import haven_utils as hu 
from .. import haven_share as hd

import os
import pprint, json
import copy
import pprint
import pandas as pd 


try:
    import ast
    from ipywidgets import Button, HBox, VBox
    from ipywidgets import widgets

    from IPython.display import display
    from IPython.core.display import Javascript, display, HTML
    from IPython.display import FileLink, FileLinks
    from ipywidgets.widgets.interaction import show_inline_matplotlib_plots
except:
    print('widgets not available...')

def share_tab(self, output):
    with output:
        l_savedir_base = widgets.Label(value="savedir_base:", layout=self.layout_label,)
        l_filterby_list= widgets.Label(value="filterby_list:", layout=self.layout_label,)
        

        bdownload = widgets.Button(description="Zip to Download Experiments", 
                                layout=self.layout_button)
                                
        bdownload_out = widgets.Output(layout=self.layout_button)

        bdownload_dropbox = widgets.Button(description="Upload to Dropbox", 
                                layout=self.layout_button)
                                
        bdownload_out_dropbox  = widgets.Output(layout=self.layout_button)

        l_fname_list = widgets.Text(
            value=str(self.vars.get('fname_list', '')),
            layout=self.layout_dropdown,
            description='fname_list:',
            disabled=False
                )

        l_dropbox_path = widgets.Text(
            value=str(self.vars.get('dropbox_path', '/shared')),
            description='dropbox_path:',
            layout=self.layout_dropdown,
            disabled=False
                )
        l_access_token_path = widgets.Text(
            value=str(self.vars.get('access_token', '')),
            description='access_token:',
            layout=self.layout_dropdown,
            disabled=False
                )
        def on_upload_clicked(b):
            fname = 'results.zip'
            bdownload_out_dropbox.clear_output()
            self.vars['fname_list'] = hu.get_list_from_str(l_fname_list.value)
            self.vars['dropbox_path'] = l_dropbox_path.value
            self.vars['access_token'] = l_access_token_path.value
            with bdownload_out_dropbox:
                self.rm.to_zip(savedir_base='', fname=fname, 
                                fname_list=self.vars['fname_list'],
                                dropbox_path=self.vars['dropbox_path'],
                                access_token=self.vars['access_token'])

            os.remove('results.zip')
            display('result.zip sent to dropbox at %s.' % self.vars['dropbox_path'])


        def on_download_clicked(b):
            fname = 'results.zip'
            bdownload_out.clear_output()
            self.vars['fname_list'] = hu.get_list_from_str(l_fname_list.value)

            with bdownload_out:
                self.rm.to_zip(savedir_base='', fname=fname, 
                                fname_list=self.vars['fname_list'])
            bdownload_out.clear_output()
            with bdownload_out:
                display('%d exps zipped.' % len(self.rm.exp_list))
            display(FileLink(fname, result_html_prefix="Download: "))
            # bdownload_out.clear_output()
            # with bdownload_out:
            #     display('%d exps zipped.' % len(self.rm.exp_list))
                
            
        bdownload.on_click(on_download_clicked)
        bdownload_zip = widgets.VBox([bdownload, bdownload_out])
        bdownload_dropbox.on_click(on_upload_clicked)
        bdownload_dropbox_vbox = widgets.VBox([ bdownload_dropbox, bdownload_out_dropbox])
        display(widgets.VBox([
                        widgets.HBox([l_savedir_base, self.t_savedir_base, ]), 
                        widgets.HBox([l_filterby_list, self.t_filterby_list,  ]),
                        widgets.HBox([l_fname_list, l_dropbox_path, l_access_token_path]),
                        widgets.HBox([bdownload_zip,
                                    bdownload_dropbox_vbox])
        ]) 
        )



import sys, os, pprint
from haven.haven_results import plots_line as pl
from haven import haven_results as hr
from haven import haven_utils as hu
from haven.haven_jobs import slurm_manager as sm

import pprint
import pandas as pd

def get_latex_table(score_df, 
                    columns=None,
                    rows=None,
                    filter_dict=dict(), 
                    map_row_dict_dict=dict(), 
                    map_col_dict=dict(), **kwargs):

    # break it
    dicts = score_df.T.to_dict()

    dicts_new = {}
    for i in dicts:
        exp_score_dict = dicts[i]
        row_label = '+'.join([str(map_row_dict_dict.get(exp_score_dict[l], exp_score_dict[l])) for l in exp_score_dict if l in rows])
        col_scores = {k:v for k, v in exp_score_dict.items() if k in columns}
        dicts_new[row_label] = col_scores

    df_new = pd.DataFrame(dicts_new).T
    return df_new.to_latex(**kwargs)